# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable
from typing import TypeVar

import pydantic

import aiopki
from aiopki.types import IAlgorithm
from aiopki.types import ISigner
from aiopki.types import IVerifier
from aiopki.utils import b64decode
from .types import JWE
from .types import JWS
from .types import JWT


T = TypeVar('T')


class JOSEObject(pydantic.RootModel[JWS | JWE]):

    @staticmethod
    def deserialize(value: str | bytes) -> dict[str, Any]:
        if isinstance(value, bytes):
            value = bytes.decode(value, 'ascii')
        params: dict[str, Any] = {}
        if value.count('.') == 2:
            protected, payload, signature = str.split(value, '.')
            params.update({
                'payload': payload,
                'signatures': [{
                    'protected': protected,
                    'signature': b64decode(signature)
                }]
            })
        if value.count('.') == 4:
            protected, key, iv, ct, aad = str.split(value, '.')
            params.update({
                'protected': protected,
                'iv': iv,
                'aad': aad,
                'ciphertext': ct,
                'recipients': [{'encrypted_key': key}]
            })
        return params

    @classmethod
    def parse_compact(cls, value: str | bytes):
        return cls.model_validate(cls.deserialize(value))

    @pydantic.model_validator(mode='before')
    def preprocess(cls, value: Any) -> Any:
        if isinstance(value, str):
            value = cls.deserialize(value)
        return value

    def payload(
        self,
        factory: Callable[[bytes], T] = JWT.model_validate
    ) -> T:
        obj = self.root
        if isinstance(obj, JWE):
            raise NotImplementedError
        assert obj.payload is not None
        return factory(b64decode(obj.payload))

    def encode(self, encoder: Callable[[bytes], T] = bytes) -> T:
        """Encodes the JWS/JWE."""
        return encoder(self.root.encode())

    async def sign(
        self,
        algorithm: str | IAlgorithm,
        signer: ISigner,
        protected: dict[str, Any] | None = None,
        header: dict[str, Any] | None = None
    ) -> None:
        """Add a signature to a JWS object."""
        if not isinstance(self.root, JWS):
            raise NotImplementedError
        if isinstance(algorithm, str):
            algorithm = aiopki.algorithm(algorithm)
        await self.root.sign(algorithm, signer, protected or {}, header or {})

    async def verify(self, verifier: IVerifier) -> bool:
        """Return a boolean indicating if at least one signature
        validated using the given verifier. Raise an exception if
        the object does not have signatures (i.e. is JWE).
        """
        return await self.root.verify(verifier)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import json

import pytest

from aiopki.ext import jose



def test_parse_jwt(jwt: str):
    obj = jose.parse(jwt)
    assert isinstance(obj.payload(), jose.JWT)


@pytest.mark.parametrize("aud", [
    ['foo'],
    'foo',
])
def test_single_audience_is_string(aud :str):
    jwt = jose.JWT.model_validate({'aud': aud})
    assert isinstance(jwt.aud, set)
    assert jwt.aud == {'foo'}


@pytest.mark.parametrize("aud", [
    ['foo'],
    'foo',
])
def test_single_audience_serializes_to_string(aud :str):
    jwt = jose.JWT.model_validate({'aud': aud})
    obj = json.loads(jwt.model_dump_json())
    assert isinstance(obj.get('aud'), str)
    assert obj.get('aud') == 'foo'


def test_has_aud():
    jwt = jose.JWT.model_validate({'aud': 'foo'})
    assert jwt.has_audience('foo')


def test_has_claims():
    jwt = jose.JWT.model_validate({'aud': 'foo'})
    assert jwt.has_claims({'aud'})


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import aiopki
from aiopki.algorithms import AESEncryption
from aiopki.algorithms import EdwardsCurveDigitalSigning
from aiopki.algorithms import EllipticCurveSigning
from aiopki.algorithms import HMACSigning
from aiopki.algorithms import RSASigning
from aiopki.types import ISigner
from aiopki.utils import b64encode
from .compactserialized import CompactSerialized
from .joseobject import JOSEObject
from .parser import parse
from .types import *


__all__: list[str] = [
    'parse',
    'sign',
    'InvalidObject',
    'JOSEObject',
    'CompactSerialized',
    'JWK',
    'JWKS',
]


def jwk(params: dict[str, Any]) -> JWK:
    return JWK.model_validate(params)


def jws(payload: dict[str, Any] | bytes, claims: dict[str, Any] | None = None) -> JOSEObject:
    claims = claims or {}
    if isinstance(payload, bytes):
        payload = b64encode(payload)
    if isinstance(payload, dict):
        claims.setdefault('typ', 'JWT')
        payload = JWT.model_validate(payload).encode()
    return JOSEObject.model_validate({
        'claims': claims,
        'payload': payload,
        'signatures': [],
    })


async def sign(
    key: ISigner,
    alg: str,
    payload: dict[str, Any] | bytes,
    protected: dict[str, Any] | None = None,
    header: dict[str, Any] | None = None,
) -> str:
    """Create a JSON Web Signature (JWS) with a JWT Claims Set or
    an unstructured payload. Return a string holding the JWS in the
    compact serialization.
    """
    obj = jws(payload, protected)
    await obj.sign(alg, key)
    return bytes.decode(obj.encode())


aiopki.register('HS256', HMACSigning('sha256'))
aiopki.register('HS384', HMACSigning('sha384'))
aiopki.register('HS512', HMACSigning('sha512'))
aiopki.register('RS256', RSASigning('sha256', 'RSASSA-PKCS1-v1_5'))
aiopki.register('ES256', EllipticCurveSigning('P-256', 'sha256'))
aiopki.register('ES384', EllipticCurveSigning('P-384', 'sha384'))
aiopki.register('ES512', EllipticCurveSigning('P-521', 'sha512'))
aiopki.register('ES256K', EllipticCurveSigning('P-256K', 'sha256'))
aiopki.register('EdDSA', EdwardsCurveDigitalSigning())

aiopki.register('A128GCM', AESEncryption(mode='GCM'))
aiopki.register('A192GCM', AESEncryption(mode='GCM'))
aiopki.register('A256GCM', AESEncryption(mode='GCM'))


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from aiopki.ext import jose


def test_jwk_thumbprint():
    jwk = jose.JWK.model_validate({
        "crv":"Ed25519",
        "kty":"OKP",
        "x":"11qYAYKxCrfVS_7TyWQHOg7hcvPapiMlrwIaaPcHURo"
    })
    assert jwk.thumbprint == 'kPrK_qmxVWaYVA9wwBF6Iuo3vVzz7TxHCTwXBygrS4k'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest
from jwcrypto import jwk
from jwcrypto import jws

from aiopki.ext import jose
from .const import SUPPORTED_KEYS


@pytest.mark.asyncio
@pytest.mark.parametrize("kid", SUPPORTED_KEYS)
async def test_sign_bytes_compact(jwks: jose.JWKS, kid: str):
    k1 = jwks.get(kid)
    assert k1.alg is not None
    assert k1.kid is not None

    k2 = jwk.JWK(**k1.model_dump(exclude_none=True)) # type: ignore
    j2 = jws.JWS(b'Hello world!')
    j2.add_signature(k2, k1.alg, {'alg': k1.alg, 'kid': k1.kid}) # type: ignore
    j1 = jose.parse(j2.serialize(compact=True))
    assert await j1.verify(k1)


@pytest.mark.asyncio
@pytest.mark.parametrize("kid", SUPPORTED_KEYS)
async def test_sign_bytes_compact_invalid(jwks: jose.JWKS, kid: str):
    """Sign and verify with a different key than that was used to sign."""
    k1 = jwks.get(kid)
    k2 = jwk.JWK(**k1.model_dump(exclude_none=True)) # type: ignore
    assert k1.alg is not None
    assert k1.kid is not None

    j2 = jws.JWS(b'Hello world!')
    j2.add_signature(k2, k1.alg, {'alg': k1.alg, 'kid': k1.kid}) # type: ignore
    j1 = jose.parse(j2.serialize(compact=True))
    assert not await j1.verify(jwks.get('example'))


@pytest.mark.asyncio
@pytest.mark.parametrize("kid", SUPPORTED_KEYS)
async def test_verify_bytes_compact(jwks: jose.JWKS, kid: str):
    k1 = jwks.get(kid)
    k2 = jwk.JWK(**k1.model_dump(exclude_none=True)) # type: ignore
    assert k1.alg is not None
    assert k1.kid is not None
    j1 = jose.jws(b'Hello world!')
    await j1.sign(k1.alg, k1, {'kid': k1.kid})

    j2 = jws.JWS()
    j2.deserialize(bytes.decode(j1.encode())) # type: ignore
    j2.verify(k2) # type: ignore


@pytest.mark.asyncio
@pytest.mark.parametrize("kid", SUPPORTED_KEYS)
async def test_verify_bytes_compact_invalid(jwks: jose.JWKS, kid: str):
    k1 = jwks.get(kid)
    k2 = jwks.get('example')
    assert k1.alg is not None
    assert k1.kid is not None
    j1 = jose.jws(b'Hello world!')
    await j1.sign(k1.alg, k1, {'kid': k1.kid})

    j2 = jws.JWS()
    j2.deserialize(bytes.decode(j1.encode())) # type: ignore
    with pytest.raises(jws.InvalidJWSSignature):
        j2.verify(k2) # type: ignore


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import urllib.parse
from typing import Any
from typing import Protocol

from canonical import HTTPResourceLocator
from canonical import ResourceName

from .icryptokey import ICryptoKey
from .isecret import ISecret


class ICryptoBackend(Protocol):
    __module__: str = 'aiopki.types'

    def handles(self, uri: urllib.parse.ParseResult) -> bool: ...
    def parse(self, name: urllib.parse.ParseResult | HTTPResourceLocator | ResourceName | str) -> ICryptoKey | None: ...
    def parse_key(self, uri: urllib.parse.ParseResult) -> ICryptoKey: ...
    def parse_secret(self, uri: urllib.parse.ParseResult) -> ISecret[Any]: ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Awaitable
from typing import Generic
from typing import TypeVar

from canonical import HTTPResourceLocator
from canonical import ResourceName
from pydantic_core import CoreSchema
from pydantic_core import core_schema
from pydantic import GetCoreSchemaHandler
from pydantic import GetJsonSchemaHandler
from pydantic.json_schema import JsonSchemaValue

from .backends import parse


__all__: list[str] = [
    'Resource'
]


S = TypeVar('S', bound='Resource[Any]')
T = TypeVar('T', bound=Awaitable[Any])


class Resource(Generic[T]):
    __module__: str = 'aiopki'
    model: type[T]

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _: Any,
        handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        return core_schema.json_or_python_schema(
            json_schema=core_schema.str_schema(max_length=2048),
            python_schema=core_schema.union_schema([
                core_schema.chain_schema([
                    core_schema.is_instance_schema(str),
                    core_schema.union_schema([
                        core_schema.no_info_plain_validator_function(ResourceName),
                        core_schema.no_info_plain_validator_function(HTTPResourceLocator),
                    ]),
                    core_schema.no_info_plain_validator_function(parse),
                    core_schema.no_info_plain_validator_function(cls),
                ]),
                core_schema.chain_schema([
                    core_schema.is_instance_schema(cls.model),
                    core_schema.no_info_plain_validator_function(cls)
                ]),
                core_schema.is_instance_schema(cls)
            ]),
            serialization=core_schema.plain_serializer_function_ser_schema(str)
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        _: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        return handler(core_schema.str_schema())

    @classmethod
    def parse(cls: type[S], name: HTTPResourceLocator | ResourceName | str) -> S:
        return cls(parse(name))

    def __init__(self, impl: T):
        self.impl = impl
        self.ready = False

    async def discover(self) -> Any:
        raise NotImplementedError

    def __await__(self):
        return self.discover().__await__()


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
import json
import operator
import secrets
from typing import Any
from typing import Callable
from typing import Iterable

import pydantic
from canonical import UnixTimestamp

from aiopki.utils import b64encode_json
from .audiencetype import AudienceType


class BaseJWT(pydantic.BaseModel):
    claims: dict[str, Any] = pydantic.Field(default={}, alias='__claims__')

    @pydantic.model_validator(mode='before')
    def preprocess(cls, values: bytes | str | dict[str, Any]) -> dict[str, Any]:
        if isinstance(values, (bytes, str)):
            values = json.loads(values)
        assert isinstance(values, dict)
        return {
            **values,
            '__claims__': {
                k: v
                for k, v in values.items()
                if k not in cls.model_fields
            }
        }
    
    def model_dump(self, **kwargs: Any) -> dict[str, Any]:
        kwargs.setdefault('exclude_none', True)
        kwargs.setdefault('exclude_defaults', True)
        kwargs.setdefault('exclude_unset', True)
        exclude = kwargs.setdefault('exclude', set())
        exclude.add('claims')
        return {**self.claims, **super().model_dump(**kwargs)}

    def encode(self, **kwargs: Any) -> bytes:
        """Encode the payload of the JWT."""
        kwargs.setdefault('exclude_defaults', True)
        kwargs.setdefault('exclude_none', True)
        return b64encode_json({
            **self.claims,
            **self.model_dump(**kwargs)
        })
    
    def has_claims(self, claims: Iterable[str]) -> bool:
        """Return a boolean indicating if the token has the given claims."""
        return set(claims) <= set(self.model_dump(exclude_none=True))


class JWT(BaseJWT):
    aud: AudienceType | None = None
    exp: UnixTimestamp | None = None
    iat: UnixTimestamp | None = None
    iss: str | None = None
    jti: str | None = None
    nbf: UnixTimestamp | None = None
    sub: str | None = None
    _now: datetime.datetime = pydantic.PrivateAttr()

    def model_post_init(self, _: Any) -> None:
        self._now = datetime.datetime.now(datetime.timezone.utc)

    def has_audience(self, value: str) -> bool:
        return self.aud is not None and value in self.aud

    def validate_aud(self, aud: str | set[str]):
        return self.aud is not None and any([
            isinstance(aud, str) and aud in self.aud,
            isinstance(aud, set) and  bool(aud & self.aud),
        ])

    def validate_exp(self, now: datetime.datetime | None = None):
        return self._validate_timestamp(self.exp, operator.ge, now)

    def validate_iat(self, max_age: int, now: datetime.datetime | None = None):
        if self.iat is None:
            return None
        exp = (now or self._now) - datetime.timedelta(seconds=max_age)
        return self.iat > exp

    def validate_iss(self, iss: str) -> bool:
        return all([
            self.iss is not None,
            self.iss == iss
        ])

    def validate_nbf(self, now: datetime.datetime | None = None):
        return self._validate_timestamp(self.nbf, operator.le, now)

    def _validate_timestamp(
        self,
        value: datetime.datetime | None,
        op: Callable[[datetime.datetime, datetime.datetime], bool],
        now: datetime.datetime | None = None
    ) -> bool:
        if value is None:
            return False
        return op(value, (now or self._now))

    @classmethod
    def new(cls, ttl: int, **kwargs: Any):
        """Create a new :class:`JWT`."""
        now = datetime.datetime.now(datetime.timezone.utc)
        return cls.model_validate({
            'jti': secrets.token_urlsafe(48),
            **kwargs,
            'exp': (now + datetime.timedelta(seconds=ttl)),
            'iat': now,
            'nbf': now,
        })


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import copy

from cryptography.hazmat.primitives.asymmetric import ec

from ..types import IAlgorithm
from .aesencryption import AESEncryption
from .edwardscurvedigitalsigning import EdwardsCurveDigitalSigning
from .ellipticcurvesigning import EllipticCurveSigning
from .hmacsigning import HMACSigning
from .notimplemented import NotImplementedAlgorithm
from .rsasigning import RSASigning


__all__: list[str] = [
    'get',
    'register',
    'AESEncryption',
    'EdwardsCurveDigitalSigning',
    'EllipticCurveSigning',
    'HMACSigning',
    'RSASigning',
]


_REGISTRY: dict[str, IAlgorithm] = {}

CURVE_MAPPING: dict[str, type[ec.EllipticCurve]] = {
    "P-256": ec.SECP256R1,
    "P-384": ec.SECP384R1,
    "P-521": ec.SECP521R1,
    "P-256K": ec.SECP256K1,
    "P-512": ec.SECP521R1,
    ec.SECP256R1.__name__: ec.SECP256R1, 
    ec.SECP384R1.__name__: ec.SECP384R1, 
    ec.SECP521R1.__name__: ec.SECP521R1, 
    ec.SECP256K1.__name__: ec.SECP256K1, 
}


def curve(name: str) -> ec.EllipticCurve:
    return CURVE_MAPPING[name]()


def register(name: str, alg: IAlgorithm) -> None:
    alg = copy.deepcopy(alg)
    alg.name = name
    _REGISTRY[name] = alg


def get(name: str) -> IAlgorithm:
    return _REGISTRY[name]


register('notimplemented', NotImplementedAlgorithm())


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import hashlib
from typing import Any
from typing import Literal

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives.asymmetric import rsa

import aiopki
from aiopki.algorithms import RSASigning
from aiopki.types import IAlgorithm
from aiopki.utils import b64decode_int
from aiopki.utils import b64encode
from aiopki.utils import b64encode_int
from .basejwk import BaseJWK


class RSAPublicKey(BaseJWK):
    kty: Literal['RSA']
    n: str
    e: str
    _public_numbers: rsa.RSAPublicNumbers

    @property
    def public(self): # pragma: no cover
        return self

    @property
    def thumbprint(self) -> str:
        message = self.public.model_dump_json(include={'e', 'kty', 'n'})
        return b64encode(hashlib.sha256(str.encode(message)).digest(), encoder=bytes.decode)


    @classmethod
    def preprocess_key(cls, key: rsa.RSAPublicKey | Any, **values: Any) -> dict[str, Any]:
        if not isinstance(key, rsa.RSAPublicKey):
            raise NotImplementedError
        numbers = key.public_numbers()
        return {
            'kty': 'RSA',
            'n': b64encode_int(numbers.n),
            'e': b64encode_int(numbers.e),
        }

    def can_use(self, algorithm: IAlgorithm) -> bool:
        return isinstance(algorithm, RSASigning)

    def default_algorithm(self) -> IAlgorithm:
        assert self.use is not None
        algorithm: IAlgorithm = aiopki.algorithms.get('notimplemented')
        match self.use:
            case 'sig':
                algorithm = aiopki.algorithms.get('RS256')
            case 'enc':
                algorithm = aiopki.algorithms.get('RSA-OAEP-256')
        return algorithm

    def is_public(self) -> bool:
        return True

    def model_post_init(self, *args: Any, **kwargs: Any) -> None:
        self._public_numbers = rsa.RSAPublicNumbers(
            e=b64decode_int(self.e),
            n=b64decode_int(self.n)
        )
        if not self.key_ops:
            self.key_ops = []
            match self.use:
                case 'sig':
                    self.key_ops.append('verify')
                    if not self.is_public():
                        self.key_ops.append('sign')
                case 'enc':
                    self.key_ops.append('wrapKey')
                    if not self.is_public():
                        self.key_ops.append('unWrapKey')
                case None:
                    self.key_ops.extend([
                        'verify',
                        'wrapKey',
                    ])
                    if not self.is_public():
                        self.key_ops.extend([
                            'sign',
                            'unWrapKey'
                        ])

    async def verify(
        self,
        signature: bytes,
        message: bytes,
        algorithm: IAlgorithm,
        using: str | None = None
    ) -> bool:
        public = self._public_numbers.public_key()
        try:
            public.verify(
                signature,
                message,
                padding=algorithm.get_padding(),
                algorithm=algorithm.get_digest_algorithm()
            )
            return True
        except InvalidSignature:
            return False


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic
import pytest

from aiopki.ext import jose

A1 = (
    'eyJ0eXAiOiJKV1QiLA0KICJhbGciOiJIUzI1NiJ9.eyJpc3MiOiJqb2UiLA0KICJl'
    'eHAiOjEzMDA4MTkzODAsDQogImh0dHA6Ly9leGFtcGxlLmNvbS9pc19yb290Ijp0c'
    'nVlfQ.dBjftJeZ4CVP-mB92K27uhbUJU1p1r_wW1gFWFOEjXk'
)

A2 = (
    'eyJhbGciOiJSUzI1NiJ9.eyJpc3MiOiJqb2UiLA0KICJleHAiOjEzMDA4MTkzODAs'
    'DQogImh0dHA6Ly9leGFtcGxlLmNvbS9pc19yb290Ijp0cnVlfQ.cC4hiUPoj9Eetd'
    'gtv3hF80EGrhuB__dzERat0XF9g2VtQgr9PJbu3XOiZj5RZmh7AAuHIm4Bh-0Qc_l'
    'F5YKt_O8W2Fp5jujGbds9uJdbF9CUAr7t1dnZcAcQjbKBYNX4BAynRFdiuB--f_nZ'
    'LgrnbyTyWzO75vRK5h6xBArLIARNPvkSjtQBMHlb1L07Qe7K0GarZRmB_eSN9383L'
    'cOLn6_dO--xi12jzDwusC-eOkHWEsqtFZESc6BfI7noOPqvhJ1phCnvWh6IeYI2w9'
    'QOYEUipUTI8np6LbgGY9Fs98rqVt5AXLIhWkWywlVmtVrBp0igcN_IoypGlUPQGe7'
    '7Rw'
)


@pytest.mark.asyncio
@pytest.mark.parametrize("value,kid", [
    (A1, 'rfc7515:a1'),
    (A2, 'rfc7515:a2')
])
async def test_parse_rfc(value: str, kid: str, jwks: jose.JWKS):
    jws = jose.parse(value)
    assert await jws.verify(jwks.get(kid))


def test_parse_bytes():
    jose.parse(str.encode(A1))


@pytest.mark.parametrize("value", [
    "foo",
    "foo.bar.baz",
    "foo..baz",
    "foo.bar.",
    ".bar.baz",
    "...",
])
def test_parse_invalid(value: str):
    with pytest.raises(pydantic.ValidationError):
        jose.parse(value)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest_asyncio

import aiopki
from canonical import ResourceName


@pytest_asyncio.fixture(scope='module') # type: ignore
async def signing_key(key_uri: ResourceName | str) -> aiopki.types.ICryptoKey:
    return await aiopki.CryptoKeyType.parse(key_uri) # type: ignore


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from aiopki.types import Base64
from .baseencryptionresult import BaseEncryptionResult


class AESEncryptionResult(BaseEncryptionResult):
    """The result of an encryption operation using the AES algorithm."""
    aad: Base64 = pydantic.Field(
        default=b''
    )

    ct: Base64 = pydantic.Field(
        default=...
    )

    iv: Base64 = pydantic.Field(
        default=...
    )
    tag: Base64 = pydantic.Field(
        default=...
    )

    def __bytes__(self) -> bytes:
        return self.ct


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio_atexit
from azure.identity.aio import DefaultAzureCredential
from azure.keyvault.keys.aio import KeyClient


credential: DefaultAzureCredential | None = None
key_clients: dict[str, KeyClient] = {}


async def get_credential() -> DefaultAzureCredential:
    global credential
    if credential is None:
        credential = DefaultAzureCredential()
        await credential.__aenter__()
        asyncio_atexit.register(close) # type: ignore
    return credential


async def get_key_client(vault_url: str) -> KeyClient:
    global key_clients
    if vault_url not in key_clients:
        key_clients[vault_url] = client = KeyClient(
            vault_url=vault_url,
            credential=await get_credential()
        )
        await client.__aenter__()
    return key_clients[vault_url]


async def close():
    global credential
    global key_clients
    for client in key_clients.values():
        await client.__aexit__(None, None, None)
    if credential is not None:
        await credential.close()




# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest

import aiopki
from aiopki.ext import jose


def test_jwks_public(jwks: jose.JWKS):
    assert all([x.is_public() for x in jwks.public.keys])


@pytest.mark.asyncio
async def test_jwks_verify(jwks: jose.JWKS, jwt: str):
    obj = jose.parse(jwt)
    assert await obj.verify(jwks)


@pytest.mark.asyncio
async def test_empty_jwks_does_not_verify(jwks: jose.JWKS, jwt: str):
    obj = jose.parse(jwt)
    mock = jose.JWKS()
    assert not await obj.verify(mock)


@pytest.mark.asyncio
async def test_unknown_kid_does_not_verify(jwks: jose.JWKS, signer: aiopki.types.ISigner):
    jws = jose.jws(b'Hello world!')
    await jws.sign(aiopki.algorithms.get('ES256'), signer, {'kid': 'foo'})
    assert not await jws.verify(jose.JWKS())


@pytest.mark.asyncio
async def test_known_kid_does_verify(jwks: jose.JWKS, signer: aiopki.types.ISigner):
    jws = jose.jws(b'Hello world!')
    await jws.sign(aiopki.algorithms.get('ES256'), signer, {'kid': 'rfc7515:a3'})
    assert await jws.verify(jwks)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import pathlib
from typing import Any

import pytest
import yaml

import aiopki
from aiopki.ext import jose
from aiopki.ext.jose import JWKS


aiopki.install_backend('aiopki.ext.cryptography')


@pytest.fixture(scope='session')
def jwks_yaml():
    with open(pathlib.Path(__file__).parent.joinpath('jwks.yml'), 'r') as f:
        return yaml.safe_load(f.read())


@pytest.fixture(scope='function')
def jwks(jwks_yaml: dict[str, Any]):
    return JWKS.model_validate(jwks_yaml)


@pytest.fixture
def jws() -> jose.JOSEObject:
    return jose.jws(b'Hello world!')


@pytest.fixture(scope="session")
def event_loop():
    loop = asyncio.new_event_loop()
    yield loop
    loop.close()


@pytest.fixture
def signer(jwks: jose.JWKS) -> jose.JWK:
    return jwks.get('rfc7515:a3')


@pytest.fixture
def untrusted_signer(jwks: jose.JWKS) -> jose.JWK:
    return jwks.get('example')


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
from typing import Union

import pydantic

import aiopki
from aiopki.types import EncryptionResult
from aiopki.types import IAlgorithm
from aiopki.types import Plaintext
from .edwardscurveprivatekey import EdwardsCurvePrivateKey
from .edwardscurvepublickey import EdwardsCurvePublicKey
from .ellipticcurveprivatekey import EllipticCurvePrivateKey
from .ellipticcurvepublickey import EllipticCurvePublicKey
from .octetprivatekey import OctetPrivateKey
from .rsaprivatekey import RSAPrivateKey
from .rsapublickey import RSAPublicKey


JWKType = Union[
    EdwardsCurvePrivateKey,
    EllipticCurvePrivateKey,
    OctetPrivateKey,
    RSAPrivateKey,
    EdwardsCurvePublicKey,
    EllipticCurvePublicKey,
    RSAPublicKey
]


class JWK(pydantic.RootModel[JWKType]):

    @property
    def alg(self) -> str | None:
        return self.root.alg

    @property
    def crv(self) -> str | None:
        return self.root.crv

    @property
    def kid(self) -> str | None:
        return self.root.kid

    @property
    def public(self) -> Union['JWK', None]:
        if self.root.public:
            return JWK.model_validate(self.root.public.model_dump())

    @property
    def thumbprint(self) -> str:
        return self.root.thumbprint

    def can_use(self, algorithm: IAlgorithm) -> bool:
        return self.root.can_use(algorithm)

    def default_algorithm(self) -> IAlgorithm:
        if self.root.alg is not None:
            return aiopki.algorithms.get(self.root.alg)
        if not self.root.use:
            raise ValueError(
                "Can not determine default algorithm if the `use` "
                "claim is not specified."
            )
        return self.root.default_algorithm()

    def is_active(self) -> bool:
        if self.root.nbf is None:
            return True
        now = datetime.datetime.now(datetime.timezone.utc).timestamp()
        return self.root.nbf <= now

    def is_available(self) -> bool:
        return not self.is_expired() and self.is_active()

    def is_expired(self) -> bool:
        if self.root.exp is None:
            return False
        now = datetime.datetime.now(datetime.timezone.utc).timestamp()
        return now >= self.root.exp

    def is_public(self) -> bool:
        return isinstance(self.root, (
            EdwardsCurvePublicKey,
            EllipticCurvePublicKey,
            RSAPublicKey,
        ))

    async def decrypt(
        self,
        ct: EncryptionResult,
        algorithm: IAlgorithm | None = None
    ) -> Plaintext:
        return await self.root.decrypt(ct, algorithm)

    async def encrypt(
        self,
        pt: bytes | Plaintext,
        algorithm: IAlgorithm | None = None
    ) -> EncryptionResult:
        return await self.root.encrypt(pt, algorithm)

    async def sign(
        self,
        message: bytes,
        algorithm: IAlgorithm | None = None,
        using: str | None = None
    ) -> bytes:
        if algorithm is None:
            algorithm = self.root.get_algorithm()
        return await self.root.sign(message, algorithm) # type: ignore

    async def verify(
        self,
        signature: bytes,
        message: bytes,
        algorithm: IAlgorithm | None = None,
        using: str | None = None
    ) -> bool:
        if not self.is_available():
            return False
        if using is not None and using not in {self.kid, self.thumbprint}:
            return False
        if algorithm is None:
            algorithm = self.root.get_algorithm()
        return await self.root.verify(signature, message, algorithm, using) # type: ignore


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .base64 import Base64
from .encryptionresult import EncryptionResult
from .ialgorithm import IAlgorithm
from .icryptobackend import ICryptoBackend
from .icryptokey import ICryptoKey
from .idecrypter import IDecrypter
from .iencrypter import IEncrypter
from .isecret import ISecret
from .isigner import ISigner
from .iverifier import IVerifier
from .plaintext import Plaintext
from .subjectid import SubjectID


__all__: list[str] = [
    'Base64',
    'EncryptionResult',
    'IAlgorithm',
    'ICryptoBackend',
    'ICryptoKey',
    'IDecrypter',
    'IEncrypter',
    'ISecret',
    'ISigner',
    'IVerifier',
    'Plaintext',
    'SubjectID',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest

from aiopki.ext import keychain


@pytest.mark.asyncio
async def test_sign_using_default(keychain: keychain.Keychain):
    key = keychain.default()
    sig = await key.sign(b'Hello world!')
    assert await keychain.verify(sig, b'Hello world!', using=key.thumbprint)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import os
import urllib.parse
from typing import Iterable

import aiofiles
from cryptography.hazmat.primitives.serialization import load_pem_private_key

import aiopki
from aiopki import BaseCryptoKey
from aiopki.types import IAlgorithm
from aiopki.types import ICryptoKey
from aiopki.ext.jose import JWK
from .cryptokeyparameters import CryptoKeyParameters


class CryptoKey(BaseCryptoKey):
    params: CryptoKeyParameters
    filename: str
    _public: JWK | None = None

    @property
    def crv(self) -> str | None:
        return self.params.crv

    @property
    def public(self) -> JWK | None:
        return self._public

    @classmethod
    def parse_uri(cls, uri: urllib.parse.ParseResult) -> 'CryptoKey':
        if uri.scheme != 'file': # pragma: no cover
            raise ValueError(f"Invalid URI: {uri}")
        return cls.model_validate({
            'params': dict(urllib.parse.parse_qsl(uri.query)),
            'filename': os.path.abspath(uri.path[1:]),
        })
    
    def can_use(self, algorithm: IAlgorithm) -> bool:
        return algorithm.name == self.params.alg

    def default_algorithm(self) -> IAlgorithm:
        if not self.params.alg:
            raise ValueError("CryptoKey does not specify a default algorithm.")
        return aiopki.algorithms.get(self.params.alg)

    def get_thumbprint(self) -> str:
        assert self.public is not None
        return self.public.thumbprint

    def is_available(self) -> bool:
        return True

    def versions(self) -> Iterable[ICryptoKey]:
        return [self]

    async def load(self) -> JWK:
        async with aiofiles.open(self.filename, 'rb') as f:
            if self.params.kty in {'EC', 'OKP', 'RSA'}:
                key = load_pem_private_key(await f.read(), None)
            else:
                raise NotImplementedError
            jwk = JWK.model_validate({'key': key, **self.params.model_dump()})
            self._public = jwk.public
        return jwk

    async def sign(
        self,
        message: bytes,
        algorithm: IAlgorithm | None = None,
        using: str | None = None
    ) -> bytes:
        jwk = await self.load()
        return await jwk.sign(message, algorithm, using)

    async def verify(
        self,
        signature: bytes,
        message: bytes,
        algorithm: IAlgorithm | None = None,
        using: str | None = None
    ) -> bool:
        jwk = await self.load()
        return await jwk.verify(signature, message, algorithm, using)
    
    def __await__(self):
        async def f():
            jwk = await self.load()
            self._public = jwk.public
            return self
        return f().__await__()


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from cryptography.hazmat.primitives.asymmetric import ec

import aiopki
from aiopki.types import IAlgorithm
from aiopki.utils import b64decode_int
from aiopki.utils import b64encode_int
from aiopki.utils import normalize_ec_signature
from .ellipticcurvepublickey import EllipticCurvePublicKey


class EllipticCurvePrivateKey(EllipticCurvePublicKey):
    d: str
    _private_numbers: ec.EllipticCurvePrivateNumbers

    @property
    def key(self) -> ec.EllipticCurvePrivateKey: # type: ignore
        return self._private_numbers.private_key()

    @property
    def public(self) -> EllipticCurvePublicKey:
        return EllipticCurvePublicKey.model_validate(self.model_dump())

    @classmethod
    def preprocess_key(cls, key: ec.EllipticCurvePrivateKey | Any, **values: Any) -> dict[str, Any]:
        if not isinstance(key, ec.EllipticCurvePrivateKey):
            raise NotImplementedError
        numbers = key.private_numbers()
        return {
            'crv': EllipticCurvePublicKey.curves[key.curve.name],
            'x': b64encode_int(numbers.public_numbers.x),
            'y': b64encode_int(numbers.public_numbers.y),
            'd': b64encode_int(numbers.private_value)
        }

    def get_algorithm(self) -> IAlgorithm:
        if self.alg is None:
            raise NotImplementedError
        return aiopki.algorithms.get(self.alg) 

    def model_post_init(self, *args: Any, **kwargs: Any) -> None:
        super().model_post_init(*args, **kwargs)
        self._private_numbers = ec.EllipticCurvePrivateNumbers(
            private_value=b64decode_int(self.d),
            public_numbers=self._public_numbers
        )

    async def sign(
        self,
        message: bytes,
        algorithm: IAlgorithm,
    ) -> bytes:
        signature = self.key.sign(
            data=message,
            signature_algorithm=ec.ECDSA(algorithm.get_digest_algorithm())
        )
        return normalize_ec_signature(
            l=(self.public_key.curve.key_size + 7) // 8,
            sig=signature
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pytest
import pytest_asyncio
from canonical import ResourceName

import aiopki


@pytest.fixture(scope='module')
def secret_name() -> ResourceName:
    return ResourceName('//secretmanager.googleapis.com/projects/641438517445/secrets/test-plain/versions/2')


@pytest_asyncio.fixture(scope='module') # type: ignore
async def secret(secret_name: Any) -> aiopki.Secret[str]:
    return aiopki.StringSecret.parse(secret_name)


@pytest.mark.asyncio
async def test_secret_content(
    secret: aiopki.Secret[str]
):
    assert await secret == 'Hello world 2!'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Literal

import pydantic

import aiopki
from aiopki.types import EncryptionResult
from aiopki.types import IAlgorithm
from aiopki.types import Plaintext


KeyOp = Literal['sign', 'verify', 'encrypt', 'decrypt', 'wrapKey', 'unWrapKey', 'deriveKey', 'deriveBits']


class BaseJWK(pydantic.BaseModel):
    alg: str | None = None
    crv: str | None = None
    exp: int | None = None
    iat: int | None = None
    key_ops: list[KeyOp] | None = None
    kid: str | None = None
    nbf: int | None = None
    use: Literal['enc', 'sig'] | None = None

    @property
    def thumbprint(self) -> str:
        raise NotImplementedError

    @pydantic.model_validator(mode='before')
    def preprocess(cls, values: dict[str, Any] | Any) -> dict[str, Any]:
        key = values.get('key', None)
        if key is not None:
            try:
                values.update(cls.preprocess_key(key), **values)
            except NotImplementedError:
                pass
        return values

    @classmethod
    def preprocess_key(cls, key: Any, **values: Any) -> dict[str, Any]:
        raise NotImplementedError

    def can_use(self, algorithm: IAlgorithm) -> bool:
        raise NotImplementedError

    def default_algorithm(self) -> IAlgorithm:
        raise NotImplementedError

    def get_algorithm(self) -> IAlgorithm:
        if not self.alg:
            raise NotImplementedError
        return aiopki.algorithms.get(self.alg)

    def get_thumbprint(self) -> str:
        return self.thumbprint

    def is_public(self) -> bool:
        raise NotImplementedError

    def process_signature(self, signature: bytes) -> bytes:
        return signature

    async def decrypt(
        self,
        ct: EncryptionResult,
        algorithm: IAlgorithm | None = None
    ) -> Plaintext:
        raise NotImplementedError

    async def encrypt(
        self,
        pt: bytes | Plaintext,
        algorithm: IAlgorithm | None = None
    ) -> EncryptionResult:
        raise NotImplementedError

    async def sign(
        self,
        message: bytes,
        algorithm: IAlgorithm,
    ) -> bytes:
        raise NotImplementedError

    async def verify(
        self,
        signature: bytes,
        message: bytes,
        algorithm: IAlgorithm,
        using: str | None = None
    ) -> bool:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from google.cloud.secretmanager import ListSecretVersionsRequest
from google.cloud.secretmanager import SecretManagerServiceAsyncClient
from google.cloud.secretmanager import SecretVersion

from aiopki import VersionedSecret
from .client import get_credential


class GoogleSecret(VersionedSecret):

    async def default(self) -> Any:
        client = SecretManagerServiceAsyncClient(
            credentials=await get_credential()
        )
        response = await client.access_secret_version( # type: ignore
            name=client.secret_version_path(
                secret_version=self.default_version,
                **client.parse_secret_path(self.name),
            )
        )
        return response.payload.data

    async def discover(self) -> None:
        client = SecretManagerServiceAsyncClient(
            credentials=await get_credential()
        )
        request = ListSecretVersionsRequest(parent=self.name)
        result = await client.list_secret_versions(request) # type: ignore
        async for version in result:
            if version.state != SecretVersion.State.ENABLED:
                continue
            params = client.parse_secret_version_path(version.name)
            if not self.has_default():
                self.default_version = params['secret_version']


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from cryptography.hazmat.primitives.asymmetric.padding import PSS, PKCS1v15
from ..types import IAlgorithm

from cryptography.hazmat.primitives.hashes import Hash
from cryptography.hazmat.primitives.hashes import SHA256
from cryptography.hazmat.primitives.hashes import SHA384
from cryptography.hazmat.primitives.hashes import SHA512
from cryptography.hazmat.primitives.asymmetric.padding import PSS
from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15


DIGEST_ALGORITHMS: dict[str, type[SHA256|SHA384|SHA512]] = {
    'sha256': SHA256,
    'sha384': SHA384,
    'sha512': SHA512,
}

PADDING_ALGORITHMS: dict[str, type[PKCS1v15|PSS]] = {
    'RSASSA-PKCS1-v1_5': PKCS1v15
}


class RSASigning(IAlgorithm):
    __module__: str = 'aiopki.algorithms'
    name: str

    def __init__(self, digalg: str, padding: str):
        self._digalg = DIGEST_ALGORITHMS[digalg]
        self._padding = PADDING_ALGORITHMS[padding]

    def digest(self, value: bytes) -> bytes:
        h = Hash(self._digalg())
        h.update(value)
        return h.finalize()

    def get_digest_algorithm(self) -> SHA256 | SHA384 | SHA512:
        return self._digalg()
    
    def get_padding(self) -> PKCS1v15 | PSS:
        params = {}
        if self._padding == PSS:
            raise NotImplementedError
        return self._padding(**params) # type: ignore


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Protocol

from .ialgorithm import IAlgorithm


class ISigner(Protocol):
    __module__: str = 'aiopki.types'

    @property
    def thumbprint(self) -> str:
        ...

    def default_algorithm(self) -> IAlgorithm: ...
    async def sign(
        self,
        message: bytes,
        algorithm: IAlgorithm | None = None,
        using: str | None = None
    ) -> bytes: ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal

import pydantic

from aiopki.lib import JSONWebKey
from aiopki.types import IAlgorithm
from .cryptokeyversion import CryptoKeyVersion


class VersionedCryptoKey(pydantic.BaseModel):
    backend: str
    name: str
    default_version: Literal['__default__'] | str = '__default__'
    annotations: dict[str, int | str] = {}
    versions: list[CryptoKeyVersion] = []
    _discovered: bool = pydantic.PrivateAttr(default=False)
    _index: dict[str, CryptoKeyVersion] = pydantic.PrivateAttr(default_factory=dict)

    @property
    def latest(self) -> CryptoKeyVersion:
        return self._index[self.default_version]

    @property
    def thumbprint(self) -> str:
        return self._index[self.default_version].thumbprint

    def add_version(
        self,
        name: str,
        alg: str,
        enabled: bool = True,
        thumbprint: str | None = None,
        public: JSONWebKey | None = None
    ) -> None:
        assert thumbprint is not None
        self._index[thumbprint] = version = CryptoKeyVersion.model_validate({
            'name': name,
            'alg': alg,
            'enabled': enabled,
            'thumbprint': thumbprint if not public else public.thumbprint,
            'public': public
        })
        self.versions.append(version)

    def can_use(self, algorithm: IAlgorithm) -> bool:
        return self.latest.alg == algorithm.name

    def get(self, using: str | None) -> CryptoKeyVersion:
        if using is None:
            using = self.default_version
        return self._index[using]

    def has(self, name: str) -> bool:
        return name in self._index

    def has_default(self) -> bool:
        return self.default_version != '__default__'

    def is_discovered(self) -> bool:
        return self._discovered

    async def discover(self) -> None:
        raise NotImplementedError

    async def sign(
        self,
        version: CryptoKeyVersion,
        message: bytes,
        algorithm: IAlgorithm,
    ) -> bytes:
        raise NotImplementedError

    async def verify(
        self,
        version: CryptoKeyVersion,
        signature: bytes,
        message: bytes,
        algorithm: IAlgorithm
    ) -> bool:
        raise NotImplementedError

    async def _discover(self):
        if not self.is_discovered():
            await self.discover()
            self._discovered = True
        return self

    def __await__(self):
        return self._discover().__await__()


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Literal

import pydantic


class VersionedSecret(pydantic.BaseModel):
    _discovered: bool = pydantic.PrivateAttr(default=False)
    backend: str
    name: str
    default_version: Literal['__default__'] | str = '__default__'
    annotations: dict[str, int | str] = {}

    async def discover(self) -> None:
        raise NotImplementedError

    def has_default(self) -> bool:
        return self.default_version != '__default__'

    def is_discovered(self) -> bool:
        return self._discovered

    async def default(self) -> Any:
        raise NotImplementedError

    async def _discover(self):
        if not self.is_discovered():
            await self.discover()
            self._discovered = True
        return self

    def __await__(self):
        return self._discover().__await__()


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from aiopki.types import ICryptoKey
from .cryptokeyuri import CryptoKeyURI


class CryptoKeySpecification(pydantic.BaseModel):
    __module__: str = 'aiopki'
    default: bool = False
    name: str
    uri: CryptoKeyURI
    using: str | None = None
    _key: ICryptoKey | None = None

    async def discover(self) -> ICryptoKey:
        if self._key is None:
            self._key = await self.uri
        return self._key

    def __await__(self):
        return self.discover().__await__()


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest_asyncio

from aiopki.ext.keychain import Keychain


@pytest_asyncio.fixture(scope='session') # type: ignore
async def keychain() -> Keychain:
    return await Keychain.model_validate({
        'keys': [
            {
                'name': 'ed448',
                'uri': 'file:///pki/sig.key?kty=OKP&crv=Ed448'
            }
        ]
    })




# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest_asyncio

import aiopki
from aiopki.types import ISigner
from aiopki.ext import jose


@pytest_asyncio.fixture(scope='function') # type: ignore
async def jwt(signer: ISigner) -> str:
    jws = jose.jws({'sub': '123', 'aud': 'test'})
    await jws.sign(aiopki.algorithms.get('ES256'), signer)
    return jws.encode(encoder=bytes.decode)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Protocol

from cryptography.hazmat.primitives.ciphers.modes import GCM


class IAlgorithm(Protocol):
    __module__: str = 'aiopki.types'
    name: str

    def get_digest_algorithm(self) -> Any: ...
    def get_padding(self) -> Any: ...

    def digest(self, value: bytes) -> bytes:
        raise NotImplementedError

    def get_digest_name(self) -> str:
        raise NotImplementedError

    def get_initialization_vector(self, iv: bytes | None = None, tag: bytes | None = None) -> tuple[GCM, bytes]:
        raise NotImplementedError

    def supports_aad(self) -> bool:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .algorithms import get as algorithm
from .algorithms import curve
from .algorithms import register
from .backends import install_backend
from .backends import parse
from .basecryptokey import BaseCryptoKey
from .cryptokeyspecification import CryptoKeySpecification
from .cryptokeytype import CryptoKeyType
from .cryptokeyversion import CryptoKeyVersion
from .cryptokeyuri import CryptoKeyURI
from .secreturi import SecretURI
from .secret import Secret
from .stringsecret import StringSecret
from .versionedcryptokey import VersionedCryptoKey
from .versionedsecret import VersionedSecret
from . import types


__all__: list[str] = [
    'algorithm',
    'curve',
    'install_backend',
    'parse',
    'register',
    'types',
    'BaseCryptoKey',
    'CryptoKeySpecification',
    'CryptoKeyType',
    'CryptoKeyURI',
    'CryptoKeyVersion',
    'Secret',
    'SecretURI',
    'StringSecret',
    'VersionedCryptoKey',
    'VersionedSecret',
]


def key(uri: str):
    pass


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest

import aiopki
from aiopki.ext import jose
from aiopki.utils import b64encode_json
from aiopki.types import ISigner


def test_protected_header_must_be_a_dict():
    encoded = bytes.join(b'.', [
        b64encode_json([]),
        b'',
        b''
    ])
    with pytest.raises(ValueError):
        jose.parse(encoded)


@pytest.mark.asyncio
async def test_header_claims_must_be_disjoint(
    signer: ISigner
):
    jws = jose.jws(b'Hello world!', claims=None)
    with pytest.raises(ValueError):
        await jws.sign(aiopki.algorithms.get('ES256'), signer, {'typ': 'foo'}, {'typ': 'foo'})


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import re
from typing import Any

import pydantic


class CryptoKeyParameters(pydantic.BaseModel):
    __module__: str = 'aiopki.ext.cryptography'
    kty: str
    crv: str | None = None
    alg: str | None = None

    @pydantic.model_validator(mode='before')
    def preprocess(cls, values: dict[str, Any]) -> dict[str, Any]:
        crv = values.get('crv')
        if crv in {'P-256', 'P-256K', 'P-384', 'P-512'}:
            assert isinstance(crv, str)
            values['kty'] = 'EC'
            if not values.get('alg'):
                values['alg'] = 'ES' + re.sub(r'^P\-', '', crv)
        return values


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric.utils import Prehashed

from aiopki.types import IAlgorithm
from aiopki.utils import b64decode_int
from aiopki.utils import b64encode_int
from .rsapublickey import RSAPublicKey


class RSAPrivateKey(RSAPublicKey):
    d: str
    p: str
    q: str
    dp: str
    dq: str
    qi: str
    _private_numbers: rsa.RSAPrivateNumbers

    @property
    def public(self) -> RSAPublicKey:
        return RSAPublicKey.model_validate(self.model_dump())

    @classmethod
    def preprocess_key(cls, key: rsa.RSAPrivateKey | Any, **values: Any) -> dict[str, Any]:
        if not isinstance(key, rsa.RSAPrivateKey):
            raise NotImplementedError
        numbers = key.private_numbers()
        return {
            'n': b64encode_int(numbers.public_numbers.n),
            'e': b64encode_int(numbers.public_numbers.e),
            'd': b64encode_int(numbers.d),
            'p': b64encode_int(numbers.p),
            'q': b64encode_int(numbers.q),
            'dp': b64encode_int(numbers.dmp1),
            'dq': b64encode_int(numbers.dmq1),
            'qi': b64encode_int(numbers.iqmp),
        }

    def is_public(self) -> bool:
        return False

    def model_post_init(self, *args: Any, **kwargs: Any) -> None:
        super().model_post_init(*args, **kwargs)
        self._private_numbers = rsa.RSAPrivateNumbers(
            public_numbers=self._public_numbers,
            d=b64decode_int(self.d),
            p=b64decode_int(self.p),
            q=b64decode_int(self.q),
            dmp1=b64decode_int(self.dp),
            dmq1=b64decode_int(self.dq),
            iqmp=b64decode_int(self.qi)
        )

    async def sign(
        self,
        message: bytes,
        algorithm: IAlgorithm
    ) -> bytes:
        private = self._private_numbers.private_key()
        return private.sign(
            data=algorithm.digest(message),
            padding=algorithm.get_padding(),
            algorithm=Prehashed(algorithm.get_digest_algorithm())
        )



# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15
from cryptography.hazmat.primitives.asymmetric.padding import PSS
from cryptography.hazmat.primitives.hashes import Hash
from cryptography.hazmat.primitives.hashes import SHA256
from cryptography.hazmat.primitives.hashes import SHA384
from cryptography.hazmat.primitives.hashes import SHA512

from ..types import IAlgorithm
from .rsasigning import DIGEST_ALGORITHMS


class EllipticCurveSigning(IAlgorithm):
    __module__: str = 'aiopki.algorithms'
    name: str

    @property
    def curve(self) -> str:
        return self._curve

    def __init__(self, curve: str, digestmod: str):
        self._curve = curve
        self._algorithm = DIGEST_ALGORITHMS[digestmod]
        self._digestmod = digestmod

    def digest(self, value: bytes) -> bytes:
        h = Hash(self._algorithm())
        h.update(value)
        return h.finalize()

    def get_digest_algorithm(self) -> SHA256 | SHA384 | SHA512:
        return self._algorithm()

    def get_digest_name(self) -> str:
        return self._digestmod

    def get_padding(self) -> PKCS1v15 | PSS:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from canonical import HTTPResourceLocator

import aiopki
from aiopki.tests.signing import *


@pytest.fixture(scope='module')
def key_uri() -> HTTPResourceLocator:
    return HTTPResourceLocator.fromstring('https://python-oauthx.vault.azure.net/keys/ES256')


@pytest.fixture(scope='module')
def signing_algorithm() -> IAlgorithm:
    return aiopki.algorithms.get('ES256')


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Generator
from typing import Iterable
from typing import Protocol

from .ialgorithm import IAlgorithm


class ICryptoKey(Protocol):
    __module__: str = 'aiopki.types'

    @property
    def public(self) -> Any:
        ...

    @property
    def thumbprint(self) -> str:
        ...

    def default(self) -> 'ICryptoKey': ...
    def default_algorithm(self) -> IAlgorithm: ...
    def get_thumbprint(self) -> str: ...
    def is_available(self) -> bool: ...
    def versions(self) -> Iterable['ICryptoKey']: ...
    async def discover(self) -> 'ICryptoKey': ...

    async def sign(
        self,
        message: bytes,
        algorithm: IAlgorithm | None = None,
        using: str | None = None
    ) -> bytes:
        ...

    async def verify(
        self,
        signature: bytes,
        message: bytes,
        algorithm: IAlgorithm | None = None,
        using: str | None = None
    ) -> bool:
        ...

    def __await__(self) -> Generator[Any, Any, 'ICryptoKey']: ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Protocol

from .ialgorithm import IAlgorithm
from .encryptionresult import EncryptionResult
from .plaintext import Plaintext


class IEncrypter(Protocol):
    
    async def encrypt(
        self,
        pt: bytes | Plaintext,
        algorithm: IAlgorithm | None = None
    ) -> EncryptionResult:
        ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.


SUPPORTED_KEYS: list[str] = [
    "rfc7515:a1",
    "rfc7515:a2",
    "rfc7515:a3",
    "rfc7515:a4",
    "rfc8037:a1"
]

ENCRYPTION_KEYS: list[tuple[str, str]] = [
    ('symmetric16', 'A128GCM'),
    ('symmetric24', 'A192GCM'),
    ('symmetric32', 'A256GCM'),
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest

from aiopki.ext import jose
from aiopki.types import IAlgorithm
from aiopki.types import ISigner
from aiopki.types import IVerifier


@pytest.mark.asyncio
async def test_sign(signing_key: ISigner, signing_algorithm: IAlgorithm, verifier: IVerifier):
    message = b'Hello world'
    signature = await signing_key.sign(message, signing_algorithm)
    assert await verifier.verify(signature, b'Hello world', signing_algorithm)


@pytest.mark.asyncio
async def test_verify(signing_key: ISigner, signing_algorithm: IAlgorithm, verifier: IVerifier):
    message = b'Hello world'
    signature = await signing_key.sign(message, signing_algorithm)
    assert not await verifier.verify(signature, b'Hello underworld', signing_algorithm)


@pytest.mark.asyncio
async def test_sign_jws(
    jws: jose.JWS,
    signing_key: ISigner,
    signing_algorithm: IAlgorithm,
    verifier: IVerifier
):
    await jws.sign(signing_algorithm, signing_key)
    assert await jws.verify(verifier)


@pytest.mark.asyncio
async def test_verify_jws(
    jws: jose.JWS,
    untrusted_signer: ISigner, 
    signing_algorithm: IAlgorithm,
    verifier: IVerifier
):
    await jws.sign(untrusted_signer.default_algorithm(), untrusted_signer)
    assert not await jws.verify(verifier)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import base64
import binascii
import json
from typing import Any
from typing import Callable
from typing import TypeVar

import crcmod
from cryptography.hazmat.primitives.asymmetric import utils


T = TypeVar('T')


def bytes_to_number(value: bytes) -> int: # pragma: no cover
    return int(binascii.b2a_hex(value), 16)


def b64decode(buf: bytes | str) -> bytes:
    """Decode the given byte-sequence or string."""
    if isinstance(buf, str):
        buf = buf.encode("ascii")
    rem = len(buf) % 4
    if rem > 0:
        buf += b"=" * (4 - rem)
    return base64.urlsafe_b64decode(buf)


def b64decode_int(value: bytes | str) -> int:
    return bytes_to_number(b64decode(value))


def b64decode_json(
    buf: bytes | str,
    encoding: str = 'utf-8',
    require: type[list[Any]] | type[dict[str, Any]] | None = None
) -> dict[str, Any] | list[Any]:
    """Deserialize a Base64-encoded string or byte-sequence as JSON."""
    result = json.loads(bytes.decode(b64decode(buf), encoding))
    if not isinstance(result, (require,) if require else (dict, list)): # pragma: no cover
        raise ValueError 
    return result


def b64encode_int(
    value: int,
    encoder: Callable[[bytes], T] = bytes
) -> T: # pragma: no cover
    return b64encode(int.to_bytes(value, (value.bit_length() + 7) // 8, 'big'), encoder=encoder)


def b64encode_json(
    obj: dict[str, Any] | list[Any],
    encoder: Callable[[bytes], T] = bytes
) -> T:
    """Encode the given dictionary as JSON and return the Base64-encoded
    byte-sequence.
    """
    return b64encode(json.dumps(obj, sort_keys=True), 'utf-8', encoder=encoder)


def b64encode(
    buf: bytes | str,
    encoding: str = 'utf-8',
    encoder: Callable[[bytes], T] = bytes
) -> T:
    """Encode the given string or byte-sequence using the specified
    encoding.
    """
    if isinstance(buf, str):
        buf = str.encode(buf, encoding=encoding)
    value = base64.urlsafe_b64encode(buf).replace(b"=", b"")
    return encoder(value)


def crc32c(data: bytes) -> int:
    """
    Calculates the CRC32C checksum of the provided data.
    Args:
        data: the bytes over which the checksum should be calculated.
    Returns:
        An int representing the CRC32C checksum of the provided bytes.
    """
    crc32c_fun = crcmod.predefined.mkPredefinedCrcFun("crc-32c") # type: ignore
    return crc32c_fun(data) # type: ignore


def crc32c_check(checksum: int, data: bytes) -> bool:
    return crc32c(data) == checksum # type: ignore


def normalize_ec_signature(l: int, sig: bytes): # pragma: no cover
    r, s = utils.decode_dss_signature(sig)
    return number_to_bytes(r, l) + number_to_bytes(s, l)


def number_to_bytes(value: int, l: int) -> bytes: # pragma: no cover
    padded = str.encode("%0*x" % (2 * l, value), "ascii")
    return binascii.a2b_hex(padded)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import urllib.parse

import pytest

from aiopki.ext import azure


def test_parse_key(key_uri: urllib.parse.ParseResult, vault_url: str):
    obj = azure.CryptoKey.parse_uri(key_uri)
    assert obj.vault_url == vault_url


def test_parse_key_version(key_version_uri: urllib.parse.ParseResult, vault_url: str):
    obj = azure.CryptoKey.parse_uri(key_version_uri)
    assert obj.vault_url == vault_url
    assert obj.default is not None


@pytest.mark.parametrize("uri", [
    "https://www.example.com",
    "https://www.example.com/foo",
    "https://foo.vault.azure.net/bar/baz"
])
def test_parse_invalid_uri(uri: str):
    with pytest.raises(ValueError):
        azure.CryptoKey.parse_uri(urllib.parse.urlparse(uri))


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Union

import pydantic

from .aesencryptionresult import AESEncryptionResult
from .dhencryptionresult import DHEncryptionResult


EncryptionResultType = Union[
    AESEncryptionResult,
    DHEncryptionResult
]
    

class EncryptionResult(pydantic.RootModel[EncryptionResultType]):

    @property
    def aad(self) -> bytes:
        return self.root.aad

    @property
    def iv(self) -> bytes | None:
        return self.root.iv

    @property
    def tag(self) -> bytes | None:
        return self.root.tag

    def __bytes__(self) -> bytes:
        return bytes(self.root)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Protocol

from .ialgorithm import IAlgorithm


class IVerifier(Protocol):
    __module__: str = 'aiopki.types'

    def can_use(self, algorithm: IAlgorithm) -> bool: ...

    async def verify(
        self,
        signature: bytes,
        message: bytes,
        algorithm: IAlgorithm,
        using: str | None = None
    ) -> bool: ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import hashlib
from typing import Any
from typing import Literal
from typing import TypeAlias
from typing import Union

import pydantic
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives.asymmetric import ed25519
from cryptography.hazmat.primitives.asymmetric import ed448

from aiopki.algorithms import EdwardsCurveDigitalSigning
from aiopki.types import IAlgorithm
from aiopki.utils import b64decode
from aiopki.utils import b64encode
from .basejwk import BaseJWK



PublicKeyType: TypeAlias = Union[
    ed448.Ed448PublicKey,
    ed25519.Ed25519PublicKey,
]


class EdwardsCurvePublicKey(BaseJWK):
    kty: Literal['OKP']
    alg: Literal['EdDSA'] = 'EdDSA' # type: ignore
    crv: Literal['Ed448', 'Ed25519'] # type: ignore
    x: str
    _public: PublicKeyType

    @property
    def public(self): # pragma: no cover
        return self

    @property
    def public_key(self) -> PublicKeyType: # pragma: no cover
        return self._public

    @property
    def thumbprint(self) -> str:
        message = self.public.model_dump_json(include={'crv', 'kty', 'x'})
        return b64encode(hashlib.sha256(str.encode(message)).digest(), encoder=bytes.decode)

    @pydantic.model_validator(mode='before')
    def preprocess_curve(cls, values: dict[str, Any]) -> dict[str, Any]:
        crv = values.get('crv')
        if crv in {'Ed448', 'Ed25519'}:
            values.update({'kty': 'OKP', 'alg': 'EdDSA', 'use': 'sig'})
        return values

    def can_use(self, algorithm: IAlgorithm) -> bool:
        return isinstance(algorithm, EdwardsCurveDigitalSigning)

    def get_algorithm(self) -> IAlgorithm:
        return EdwardsCurveDigitalSigning(self.alg)

    def model_post_init(self, *args: Any, **kwargs: Any) -> None:
        self.key_ops = ['verify']
        match self.crv:
            case 'Ed448':
                cls = ed448.Ed448PublicKey
            case 'Ed25519':
                cls = ed25519.Ed25519PublicKey
            case _: # type: ignore pragma: no cover
                raise NotImplementedError
        self._public = cls.from_public_bytes(b64decode(self.x))

    async def verify(
        self,
        signature: bytes,
        message: bytes,
        algorithm: IAlgorithm,
        using: str | None = None
    ) -> bool:
        if not self.can_use(algorithm):
            raise NotImplementedError(
                f"{type(self).__name__} does not support signature verification "
                f"using the {algorithm.name} algorithm."
            )
        try:
            self._public.verify(signature, message)
            return True
        except InvalidSignature:
            return False



# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools

from .joseobject import JOSEObject



@functools.singledispatch
def parse(value: bytes | str) -> JOSEObject:
    raise NotImplementedError(type(value).__name__)


@parse.register
def parse_compact(value: str) -> JOSEObject:
    return JOSEObject.parse_compact(value)


@parse.register
def parse_compact_bytes(value: bytes) -> JOSEObject:
    return JOSEObject.parse_compact(value)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pydantic

from aiopki.types import IAlgorithm
from aiopki.types import ISigner
from aiopki.types import IVerifier
from .signature import Signature


class JWS(pydantic.BaseModel):
    payload: bytes
    signatures: list[Signature] = []
    claims: dict[str, Any] = {}

    def encode(self, compact: bool = True) -> bytes:
        """Encodes the JWS/JWE."""
        if len(self.signatures) > 1: # pragma: no cover
            raise ValueError("Compact encoding can not be used with multiple signatures.")
        return self.encode_compact()

    def encode_compact(self) -> bytes:
        assert len(self.signatures) == 1
        return self.signatures[0].encode(self.payload)

    async def sign(
        self,
        algorithm: IAlgorithm,
        signer: ISigner,
        protected: dict[str, Any] | None = None,
        header: dict[str, Any] | None = None
    ) -> Signature:
        sig = await Signature.new(algorithm, signer, self.payload, protected=protected, header=header)
        self.signatures.append(sig)
        return sig

    async def verify(self, verifier: IVerifier) -> bool:
        """Return a boolean indicating if at least one signature
        validated using the given verifier.
        """
        for signature in self.signatures:
            is_valid = await signature.verify(verifier, self.payload)
            if not is_valid:
                continue
            break
        else:
            is_valid = False
        return is_valid


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from canonical import ResourceName

import aiopki
from aiopki.tests.signing import *


@pytest.fixture(scope='module')
def key_uri() -> ResourceName:
    return ResourceName('//cloudkms.googleapis.com/projects/unimatrixdev/locations/europe-west4/keyRings/local/cryptoKeys/hmac_sha256')

@pytest.fixture
def signing_algorithm() -> IAlgorithm:
    return aiopki.algorithms.get('HS256')


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .audiencetype import AudienceType
from .jwa import JWA
from .jwe import JWE
from .jwk import JWK
from .jwks import JWKS
from .jws import JWS
from .jwsheader import JWSHeader
from .jwt import BaseJWT
from .jwt import JWT
from .oidctoken import OIDCToken


__all__: list[str] = [
    'AudienceType',
    'BaseJWT',
    'InvalidObject',
    'JWA',
    'JWE',
    'JWK',
    'JWKS',
    'JWS',
    'JWSHeader',
    'JWT',
    'OIDCToken',
]


class InvalidObject(TypeError):
    pass


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Iterable
from typing import TypeVar

import pydantic

from aiopki.types import IAlgorithm
from aiopki.types import ICryptoKey


T = TypeVar('T', bound='BaseCryptoKey')


class BaseCryptoKey(pydantic.BaseModel):

    @property
    def thumbprint(self) -> str:
        return self.get_thumbprint()

    @property
    def public(self) -> Any:
        raise NotImplementedError
    
    def can_use(self, algorithm: IAlgorithm) -> bool:
        raise NotImplementedError

    def default(self) -> ICryptoKey:
        return self

    def default_algorithm(self) -> IAlgorithm:
        raise NotImplementedError

    def get_thumbprint(self) -> str:
        raise NotImplementedError

    def is_available(self) -> bool:
        raise NotImplementedError

    async def _discover(self: T) -> T:
        return await self.discover()
    
    async def discover(self):
        return self

    async def sign(
        self,
        message: bytes,
        algorithm: IAlgorithm | None = None,
        using: str | None = None
    ) -> bytes:
        raise NotImplementedError

    async def verify(
        self,
        signature: bytes,
        message: bytes,
        algorithm: IAlgorithm | None = None,
        using: str | None = None
    ) -> bool:
        raise NotImplementedError

    def versions(self) -> Iterable[ICryptoKey]:
        raise NotImplementedError

    def __await__(self):
        return self._discover().__await__()


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic
import pytest

import aiopki
from aiopki import backends


SUPPORTED_URIS: list[str] = [
    "file:///pki/sig.key?kty=OKP&alg=EdDSA&crv=Ed448",
    #"file:///pki/rsa.key?kty=RSA",
    "file:///pki/secp256k1.key?crv=P-256K",
    "file:///pki/secp384r1.key?crv=P-384",
    "file:///pki/secp521r1.key?crv=P-512",
]


class Model(pydantic.BaseModel):
    key: aiopki.CryptoKeyURI


@pytest.mark.asyncio
@pytest.mark.parametrize("uri", SUPPORTED_URIS)
async def test_parse_uri_signing_alg(uri: str):
    k = backends.parse_uri(uri)
    sig = await k.sign(b'Hello world')
    assert await k.verify(sig, b'Hello world')


@pytest.mark.parametrize("uri", SUPPORTED_URIS)
@pytest.mark.asyncio
async def test_parse_uri_on_pydantic_model(uri: str):
    obj = Model.model_validate({
        'key': uri
    })
    key = await obj.key
    sig = await key.sign(b'Hello world!')
    assert await key.verify(sig, b'Hello world!')


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pydantic

from aiopki.types import EncryptionResult
from aiopki.types import IAlgorithm
from aiopki.types import Plaintext
from .jwk import JWK


class JWKS(pydantic.BaseModel):
    keys: list[JWK] = []
    _index: dict[str, JWK] = {}

    @property
    def crv(self) -> str | None:
        return self.keys[0].crv

    @property
    def public(self) -> 'JWKS':
        return JWKS.model_validate({
            'keys': filter(bool, [x.public for x in self.keys])
        })

    @property
    def thumbprint(self) -> str:
        return self.keys[0].thumbprint

    def can_use(self, algorithm: IAlgorithm) -> bool:
        return any([k.can_use(algorithm) for k in self.keys])

    def default(self) -> JWK:
        return self.keys[0]

    def default_algorithm(self) -> IAlgorithm:
        return self.keys[0].default_algorithm()

    def get(self, kid: str) -> JWK:
        return self._index[kid]

    def model_post_init(self, *args: Any, **kwargs: Any) -> None:
        for jwk in self.keys:
            self._index[jwk.thumbprint] = jwk
            if jwk.kid is not None:
                self._index[jwk.kid] = jwk

    async def decrypt(
        self,
        ct: EncryptionResult,
        algorithm: IAlgorithm | None = None,
        using: str | None = None
    ) -> Plaintext:
        keys = [x for x in self.keys if x.root.use == 'enc']
        if not keys:
            raise NotImplementedError
        if using:
            key = self.get(using)
        else:
            key = keys[0]
        return await key.decrypt(ct, algorithm)

    async def encrypt(
        self,
        pt: bytes | Plaintext,
        algorithm: IAlgorithm | None = None,
        using: str | None = None
    ) -> EncryptionResult:
        keys = [x for x in self.keys if x.root.use == 'enc']
        if not keys:
            raise NotImplementedError
        if using:
            key = self.get(using)
        else:
            key = keys[0]
        return await key.encrypt(pt, algorithm)

    async def sign(
        self,
        message: bytes,
        algorithm: IAlgorithm | None = None,
        using: str | None = None
    ) -> bytes:
        return await (self.get(using) if using else self.default()).sign(
            message=message,
            algorithm=algorithm,
            using=using
        )

    async def verify(
        self,
        signature: bytes,
        message: bytes,
        algorithm: IAlgorithm,
        using: str | None = None
    ) -> bool:
        """Verify the given signature and return a boolean indicating
        if the signature is valid.
        
        If the `using` parameter is not ``None``, then it must point
        to a known `kid` in the JWKS. An unknown value in the `kid`
        parameter immediately returns ``False``. Otherwise, all keys
        in the JWKS are attempted.
        """
        if using is not None and using not in self._index:
            return False

        if using is not None:
            return await self._index[using].verify(signature, message, algorithm)

        return any([
            await k.verify(signature, message, algorithm)
            for k in self.keys if k.can_use(algorithm)
        ])


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .types import IAlgorithm
from .resource import Resource
from .versionedcryptokey import VersionedCryptoKey


__all__: list[str] = [
    'CryptoKeyType'
]


class CryptoKeyType(Resource[VersionedCryptoKey]):
    __module__: str = 'aiopki'
    model = VersionedCryptoKey

    @property
    def crv(self) -> str | None:
        return None

    @property
    def thumbprint(self) -> str:
        return self.impl.thumbprint

    def can_use(self, algorithm: IAlgorithm) -> bool:
        return self.impl.can_use(algorithm)

    async def discover(self):
        if not self.ready:
            await self.impl
        return self

    async def sign(
        self,
        message: bytes,
        algorithm: IAlgorithm | None = None,
        using: str | None = None
    ) -> bytes:
        await self.impl
        version = self.impl.get(using)
        return await self.impl.sign(version, message, algorithm or version.default_algorithm())

    async def verify(
        self,
        signature: bytes,
        message: bytes,
        algorithm: IAlgorithm | None = None,
        using: str | None = None
    ) -> bool:
        await self.impl
        if using and not self.impl.has(using):
            return False
        version = self.impl.get(using)
        return await self.impl.verify(
            version=version,
            signature=signature,
            message=message,
            algorithm=algorithm or version.default_algorithm(),
        )

    def __await__(self):
        return self.discover().__await__()


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from .secret import Secret


__all__: list[str] = [
    'StringSecret'
]


class StringSecret(Secret[str]):
    __module__: str = 'aiopki'

    async def decode(self, value: Any) -> str:
        return bytes.decode(value)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from cryptography.hazmat.primitives.asymmetric.padding import PSS, PKCS1v15
from cryptography.hazmat.primitives.hashes import SHA256, SHA384, SHA512
from ..types import IAlgorithm


class EdwardsCurveDigitalSigning(IAlgorithm):
    __module__: str = 'aiopki.algorithms'
    name: str
    curve: str | None

    def __init__(self, curve: str | None = None) -> None:
        self.curve = curve

    def get_digest_algorithm(self) -> SHA256 | SHA384 | SHA512:
        raise NotImplementedError

    def get_padding(self) -> PKCS1v15 | PSS:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import os
from typing import Literal

from cryptography.hazmat.primitives.ciphers.modes import CBC
from cryptography.hazmat.primitives.ciphers.modes import GCM

from .encryptionalgorithm import EncryptionAlgorithm


class AESEncryption(EncryptionAlgorithm):
    __module__: str = 'oauthx.algorithms'
    iv_length: int = 12
    _modes: dict[str, type[CBC | GCM]] = {
        'CBC': CBC,
        'GCM': GCM
    }

    def __init__(self, mode: Literal['CBC', 'GCM']) -> None:
        self._mode = mode

    def get_initialization_vector(self, iv: bytes | None = None, tag: bytes | None = None) -> tuple[GCM, bytes]:
        Mode = self._modes[self._mode]
        iv = iv or os.urandom(self.iv_length)
        
        assert Mode == GCM
        return Mode(iv, tag=tag), iv # type: ignore

    def supports_aad(self) -> bool:
        return self._mode == 'GCM'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pydantic

from aiopki.utils import b64encode
from .jwa import JWA
from .jwk import JWK


class JWSHeader(pydantic.BaseModel):
    alg: JWA
    crv: str | None = None
    cty: str | None = None
    crit: list[str] = []
    jku: str | None = None
    jwk: JWK | None = None
    kid: str | None = None
    typ: str | None = None
    x5c: str | None = None
    x5t: str | None = None
    x5t_sha256: str | None = pydantic.Field(default=None, alias='x5t#S256')
    x5u: str | None = None
    _extra: dict[str, Any]

    @pydantic.model_validator(mode='before')
    def preprocess(cls, values: dict[str, Any]) -> dict[str, Any]:
        return values

    def encode(self, **kwargs: Any) -> bytes:
        kwargs.setdefault('exclude_none', True)
        kwargs.setdefault('exclude_defaults', True)
        kwargs.setdefault('exclude_unset', True)
        return b64encode(self.model_dump_json(**kwargs))


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Union

from cryptography.hazmat.primitives.asymmetric import ed25519
from cryptography.hazmat.primitives.asymmetric import ed448

from aiopki.utils import b64decode
from aiopki.utils import b64encode
from aiopki.types import IAlgorithm
from .edwardscurvepublickey import EdwardsCurvePublicKey


PrivateKeyClass: tuple[type, ...] = (
    ed448.Ed448PrivateKey,
    ed25519.Ed25519PrivateKey
)

PrivateKeyType = Union[
    ed448.Ed448PrivateKey,
    ed25519.Ed25519PrivateKey
]


class EdwardsCurvePrivateKey(EdwardsCurvePublicKey):
    d: str

    @property
    def public(self): # pragma: no cover
        return EdwardsCurvePublicKey.model_validate(self.model_dump())

    @classmethod
    def preprocess_key(cls, key: PrivateKeyType, **values: Any) -> dict[str, Any]:
        if not isinstance(key, PrivateKeyClass):
            raise NotImplementedError
        public = key.public_key()
        return {
            'x': b64encode(public.public_bytes_raw()),
            'd': b64encode(key.private_bytes_raw())
        }
        

    def model_post_init(self, *args: Any, **kwargs: Any) -> None:
        super().model_post_init(*args, **kwargs)
        self.key_ops = ['sign', 'verify']
        match self.crv:
            case 'Ed448':
                cls = ed448.Ed448PrivateKey
            case 'Ed25519':
                cls = ed25519.Ed25519PrivateKey
            case _: # type: ignore pragma: no cover
                raise NotImplementedError
        self._private = cls.from_private_bytes(b64decode(self.d))

    async def sign(
        self,
        message: bytes,
        algorithm: IAlgorithm
    ) -> bytes:
        if not self.can_use(algorithm):
            raise NotImplementedError(
                f"{type(self).__name__} does not support signing "
                f"using the {algorithm.name} algorithm."
            )
        return self._private.sign(message)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from aiopki.lib import JSONWebKey
from aiopki.types import IAlgorithm
from .algorithms import get as algorithm


class CryptoKeyVersion(pydantic.BaseModel):
    name: str
    alg: str
    enabled: bool
    thumbprint: str
    public: JSONWebKey | None = None

    def default_algorithm(self) -> IAlgorithm:
        return algorithm(self.alg)

    def process_signature(self, signature: bytes) -> bytes:
        if self.public is not None:
            signature = self.public.process_signature(signature)
        return signature



# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import importlib
import urllib.parse
from typing import cast

from canonical import HTTPResourceLocator
from canonical import ResourceName

from .types import ICryptoBackend
from .versionedcryptokey import VersionedCryptoKey


_BACKENDS: list[ICryptoBackend] = []


def install_backend(qualname: str, _version: str = 'v1') -> None:
    """Given the qualified name to a Python module, install
    the backend implementation of cryptographic operations.
    """
    module = importlib.import_module(qualname)
    if module not in _BACKENDS: # type: ignore
        _BACKENDS.insert(0, module) # type: ignore


def get_backend(uri: str | urllib.parse.ParseResult) -> ICryptoBackend:
    install_backend('aiopki.ext.cryptography')
    if not isinstance(uri, urllib.parse.ParseResult):
        uri = urllib.parse.urlparse(uri)
    backend = None
    for candidate in _BACKENDS:
        if not candidate.handles(uri):
            continue
        backend = candidate
        break
    if backend is None:
        raise ValueError("Unable to determine backend for URI.")
    return backend


def parse_uri(uri: str):
    backend = get_backend(uri)
    return backend.parse_key(urllib.parse.urlparse(uri))


def parse(name: HTTPResourceLocator | ResourceName | str) -> VersionedCryptoKey:
    obj = None
    for candidate in _BACKENDS:
        obj = candidate.parse(name)
        if obj is not None:
            break
    if obj is None:
        raise ValueError(f"Unknown key type: {name}")
    return cast(VersionedCryptoKey, obj)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
from typing import Any

from aiopki import BaseCryptoKey
from aiopki import CryptoKeySpecification
from aiopki.ext.jose import JWKS
from aiopki.types import IAlgorithm
from aiopki.types import ICryptoKey


__all__: list[str] = [
    'Keychain'
]


class Keychain(BaseCryptoKey):
    keys: list[CryptoKeySpecification] = []
    _default: str | None = None
    _discovered: bool = False
    _index: dict[str, ICryptoKey] = {}
    _trust: dict[str, ICryptoKey] = {}

    @property
    def crv(self) -> str:
        # TODO: remove this attribute
        return self.default().crv # type: ignore

    def as_jwks(self) -> JWKS:
        keys: list[dict[str, Any]] = []
        for kid, obj in self._trust.items():
            keys.append({**obj.public.model_dump(), 'kid': kid})
        return JWKS.model_validate({'keys': keys})

    def model_post_init(self, _: Any) -> None:
        self._index = {}
        self._trust = {}

    def can_use(self, algorithm: IAlgorithm) -> bool:
        return True

    def default(self) -> ICryptoKey:
        assert self._default
        return self._index[self._default]

    def default_algorithm(self) -> IAlgorithm:
        return self.default().default_algorithm()

    def get(self, using: str) -> ICryptoKey:
        return self._index[using]

    def get_thumbprint(self) -> str:
        return self.default().get_thumbprint()

    async def discover(self):
        if not self._discovered:
            await asyncio.gather(*map(CryptoKeySpecification.discover, self.keys))
            for key in self.keys:
                self._index[key.name] = await key
                if self._default is not None and key.default:
                    raise ValueError("Multiple default keys defined.")
                if key.default:
                    self._default = key.name
                for version in self._index[key.name].versions():
                    if not version.is_available():
                        continue
                    self._trust[version.get_thumbprint()] = version
        if self._default is None:
            self._default = self.keys[0].name
        self._discovered = True
        return self

    async def sign(
        self,
        message: bytes,
        algorithm: IAlgorithm | None = None,
        using: str | None = None
    ) -> bytes:
        key = self.get(using) if using else self.default()
        return await key.sign(message, algorithm)

    async def verify(
        self,
        signature: bytes,
        message: bytes,
        algorithm: IAlgorithm | None = None,
        using: str | None = None
    ) -> bool:
        if using not in self._trust:
            return False
        return await self._trust[using].verify(signature, message, algorithm)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime

from canonical import EmailAddress
from canonical import Phonenumber

from aiopki.types import SubjectID
from .jwt import BaseJWT


class OIDCToken(BaseJWT):
    aud: str | list[str]
    exp: int
    iat: int
    iss: str
    sub: str
    auth_time: int | None = None
    nonce: str | None = None
    acr: str | None = None
    amr: str | None = None
    azp: str | None = None
    at_hash: str | None = None
    c_hash: str | None = None
    name: str | None = None
    given_name: str | None = None
    family_name: str | None = None
    middle_name: str | None = None
    nickname: str | None = None
    preferred_username: str | None = None
    profile: str | None = None
    picture: str | None = None
    website: str | None = None
    email: EmailAddress | None = None
    email_verified: bool = False
    gender: str | None = None
    birthdate: str | None = None
    zoneinfo: str | None = None
    locale: str | None = None
    phone_number: Phonenumber | str | None = None
    phone_number_verified: bool = False
    updated_at: datetime.datetime | None = None
    
    @property
    def subject_id(self) -> SubjectID:
        return SubjectID(iss=self.iss, sub=self.sub)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pydantic

import aiopki
from aiopki.types import IAlgorithm
from aiopki.types import ISigner
from aiopki.types import IVerifier
from aiopki.utils import b64decode_json
from aiopki.utils import b64encode
from .invalidobject import InvalidObject
from .jwsheader import JWSHeader


class Signature(pydantic.BaseModel):
    claims: JWSHeader
    protected: bytes | None = None
    header: dict[str, Any] = {}
    signature: bytes

    @property
    def alg(self) -> IAlgorithm:
        return aiopki.algorithm(self.claims.alg)

    @property
    def kid(self) -> str | None:
        return self.claims.kid

    @classmethod
    async def new(
        cls,
        algorithm: IAlgorithm,
        signer: ISigner,
        payload: bytes,
        protected: JWSHeader | dict[str, Any] | None = None,
        header: dict[str, Any] | None = None,
    ):
        if not isinstance(protected, JWSHeader):
            protected = JWSHeader.model_validate({
                **(protected or {}),
                'alg': algorithm.name,
                'crv': signer.crv,
                'kid': signer.thumbprint
            })
        assert isinstance(protected, JWSHeader), type(protected).__name__
        assert protected.alg == algorithm.name
        message = bytes.join(b'.', [protected.encode(), payload])
        return cls.model_validate({
            'protected': protected.encode(),
            'signature': await signer.sign(message, algorithm),
            'header': header or {}
        })


    @pydantic.field_validator('protected', mode='before')
    def preprocess_protected(cls, value: bytes | str | None) -> bytes | None:
        if isinstance(value, str):
            value = str.encode(value, 'ascii')
        return value
    
    @pydantic.model_validator(mode='before')
    def preprocess(
        cls,
        values: dict[str, Any]
    ) -> dict[str, Any]:
        # The Header Parameter values used when creating or validating
        # individual signature or MAC values are the union of the two
        # sets of Header Parameter values that may be present: (1) the
        # JWS Protected Header represented in the "protected" member of
        # the signature/MAC's array element, and (2) the JWS Unprotected
        # Header in the "header" member of the signature/MAC's array element.
        # The union of these sets of Header Parameters comprises the JOSE
        # Header.  The Header Parameter names in the two locations MUST
        # be disjoint.
        claims = values.get('header') or {}
        protected = {}
        if values.get('protected'):
            protected = b64decode_json(values['protected'])
        if not isinstance(protected, dict):
            raise InvalidObject("The encoded protected header must be a JSON object.")
        if set(claims.keys()) & set(protected.keys()):
            raise InvalidObject(
                "The header parameter names in the protected and "
                "unprotected header must be disjoint."
            )
        values['claims'] = {**claims, **protected}
        return values

    def encode(self, payload: bytes) -> bytes:
        if self.protected is None: # pragma: no cover
            raise ValueError("Missing protected header.")
        return bytes.join(b'.', [
            self.protected,
            payload,
            b64encode(self.signature)
        ])

    async def verify(self, verifier: IVerifier, payload: bytes) -> bool:
        """Return a boolean indicating if the signature
        is valid.
        """
        assert self.protected is not None
        if not verifier.can_use(self.alg):
            return False

        return await verifier.verify(
            self.signature,
            bytes.join(b'.', [self.protected, payload]),
            algorithm=self.alg,
            using=self.kid,
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic
import pytest

import aiopki
from aiopki.ext import jose


class Model(pydantic.BaseModel):
    jwt: jose.CompactSerialized



@pytest.mark.asyncio
async def test_parse_josetype(jwks: jose.JWKS):
    alg = aiopki.algorithms.get('ES256')
    jwk = jwks.get('example')
    jws = jose.jws({'msg': 'Hello world!'})
    await jws.sign(alg, jwk)

    obj = Model.model_validate({'jwt': jws.encode(encoder=bytes.decode)})
    assert isinstance(obj.jwt, jose.JOSEObject)
    assert await obj.jwt.verify(jwk)
    jwt = obj.jwt.payload()
    assert jwt.claims.get('msg') == 'Hello world!'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import re
from typing import Any

from azure.identity.aio import DefaultAzureCredential
from azure.keyvault.keys.crypto.aio import CryptographyClient

from aiopki import VersionedCryptoKey
from aiopki import CryptoKeyVersion
from aiopki.lib import JSONWebKey
from aiopki.types import IAlgorithm
from aiopki.utils import b64encode
from .client import get_credential
from .client import get_key_client


EC_ALGORITHMS = {
    'P-256': 'ES256',
    'P-384': 'ES384'
}


class AzureKey(VersionedCryptoKey):

    def model_post_init(self, _: Any) -> None:
        super().model_post_init(_)
        if self.backend != 'vault.azure.net':
            raise ValueError(f"Invalid backend: {self.backend}")

    async def discover(self):
        client = await get_key_client(str(self.annotations['vault_url']))
        async for props in client.list_properties_of_key_versions(self.name): # type: ignore
            assert props.created_on is not None
            if props.version is None:
                continue
            version = await client.get_key(props.name, version=props.version) # type: ignore
            public = None
            is_enabled = props.enabled if props.enabled is not None else True
            if str(version.key_type or '') not in {'EC', 'EC-HSM', 'RSA', 'RSA-HSM'}:
                raise NotImplementedError
            kty = re.sub(r'^(oct|EC|RSA)(\-HSM)?$', r'\1', str(version.key_type))
            public = {}
            if version.key.crv: # type: ignore
                public['alg'] = EC_ALGORITHMS[version.key.crv] # type: ignore
            for attname in version.key._FIELDS: # type: ignore
                v = getattr(version.key, attname)
                if isinstance(v, bytes):
                    v = b64encode(v)
                if attname == 'key_ops':
                    v = [str(x) for x in v]
                public[attname] = v
            public = JSONWebKey.model_validate({
                **public,
                'kty': kty
            })
            assert public.alg is not None
            self.add_version(
                name=props.version,
                alg=public.alg,
                enabled=is_enabled,
                thumbprint=public.thumbprint,
                public=public
            )
            if not self.has_default() and is_enabled:
                self.default_version = public.thumbprint

    async def sign(
        self,
        version: CryptoKeyVersion,
        message: bytes,
        algorithm: IAlgorithm,
    ) -> bytes:
        message = algorithm.digest(message)
        async with self.client(version, await get_credential()) as client:
            result = await client.sign(version.public.alg, message) # type: ignore
        return result.signature # type: ignore

    async def verify(
        self,
        version: CryptoKeyVersion,
        signature: bytes,
        message: bytes,
        algorithm: IAlgorithm
    ) -> bool:
        assert version.public is not None
        return await version.public.verify(signature, message, algorithm)

    def client(self, version: CryptoKeyVersion, credential: DefaultAzureCredential) -> CryptographyClient:
        return CryptographyClient(
            f"{self.annotations['vault_url']}/keys/{self.name}/{version.name}",
            credential=credential
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pytest
import pytest_asyncio

import aiopki


aiopki.install_backend('aiopki.ext.azure')


@pytest.fixture(scope='module')
def vault_url() -> str:
    return 'https://python-oauthx.vault.azure.net/'


@pytest_asyncio.fixture(scope='module') # type: ignore
async def verifier(signing_key: Any):
    return signing_key


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from aiopki.ext import jose


K1 = jose.jwk({
    'kty': 'oct',
    'k': 'AyM1SysPpbyDfgZld3umj1qzKObwVMkoqQ-EstJQLr_T-1qS0gZH75aKtMN3Yj0iPS4hcgUuTwjAzZr1Z9CAow'
})


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest

import aiopki
from aiopki.ext import jose
from .const import SUPPORTED_KEYS


@pytest.mark.asyncio
@pytest.mark.parametrize("kid", SUPPORTED_KEYS)
async def test_sign(jwks: jose.JWKS, kid: str):
    jwk = jwks.get(kid)
    assert jwk.alg is not None
    alg = aiopki.algorithms.get(jwk.alg)
    sig = await jwk.sign(b'Hello world!', alg)
    assert await jwk.verify(sig, b'Hello world!', alg)


@pytest.mark.asyncio
@pytest.mark.parametrize("kid", SUPPORTED_KEYS)
async def test_signed_with_different_key(jwks: jose.JWKS, kid: str):
    k = jwks.get('example')

    assert k.alg is not None
    jws = jose.jws({'kid': k.kid})
    await jws.sign(k.alg, k) # type: ignore
    assert not await jws.verify(jwks.get(kid))


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import re
import urllib.parse

from canonical import HTTPResourceLocator
from canonical import ResourceName

from aiopki.types import ICryptoKey

from .azurekey import AzureKey


def parse(name: ResourceName | HTTPResourceLocator) -> ICryptoKey | None:
    if not isinstance(name, HTTPResourceLocator):
        return None
    uri = urllib.parse.urlparse(str(name))
    if uri.hostname is None: # pragma: no cover
        return None
    if not re.match(r'^.*\.vault\.azure\.net', uri.hostname):
        return None
    m = re.match(r'^/keys/([^/]+)(/[0-9a-z]+)?$', uri.path)
    if m is None:
        return None
    key, version = m.groups()
    if version:
        version = re.sub(r'[^a-z0-9]+', '', version)
    return AzureKey.model_validate({ # type: ignore
        'backend': 'vault.azure.net',
        'name': key,
        'default_version': version or '__default__',
        'annotations': {
            'vault_url': f'{uri.scheme}://{uri.netloc}'
        }
    })


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import hashlib
import hmac
from typing import Any
from typing import Literal

from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers.algorithms import AES

import aiopki
from aiopki.algorithms import HMACSigning
from aiopki.types import EncryptionResult
from aiopki.types import IAlgorithm
from aiopki.types import Plaintext
from aiopki.utils import b64decode
from aiopki.utils import b64encode
from .basejwk import BaseJWK


class OctetPrivateKey(BaseJWK):
    k: str
    kty: Literal['oct']

    @property
    def public(self) -> None:
        return None

    @property
    def thumbprint(self) -> str:
        message = self.model_dump_json(include={'k', 'kty'})
        return b64encode(hashlib.sha256(str.encode(message)).digest(), encoder=bytes.decode)

    def can_use(self, algorithm: IAlgorithm) -> bool:
        return isinstance(algorithm, HMACSigning)

    def default_algorithm(self) -> IAlgorithm:
        assert self.use is not None
        algorithm: IAlgorithm = aiopki.algorithms.get('notimplemented')
        match self.use:
            case 'sig':
                algorithm = aiopki.algorithms.get('HS256')
            case 'enc':
                algorithm = aiopki.algorithms.get('A256GCM')
        return algorithm

    def model_post_init(self, *args: Any, **kwargs: Any) -> None:
        self._key = b64decode(self.k)
        if not self.key_ops:
            self.key_ops = []
            match self.use:
                case 'sig':
                    self.key_ops.extend([
                        'sign',
                        'verify'
                    ])
                case 'enc':
                    if self.alg in {'dir', None}:
                        self.key_ops.extend([
                            'encrypt',
                            'decrypt'
                        ])
                    if self.alg != 'dir':
                        self.key_ops.extend([
                            'unWrapKey',
                            'wrapKey'
                        ])
                case None:
                    self.key_ops.extend([
                        'sign',
                        'verify',
                        'unWrapKey',
                        'wrapKey'
                    ])


    async def decrypt(
        self,
        ct: EncryptionResult,
        algorithm: IAlgorithm | None = None
    ) -> Plaintext:
        algorithm = algorithm or self.default_algorithm()
        mode, _ = algorithm.get_initialization_vector(ct.root.iv, ct.root.tag)
        cipher = Cipher(AES(self._key), mode)
        decrypter = cipher.decryptor()
        if algorithm.supports_aad():
            decrypter.authenticate_additional_data(ct.aad)
        return Plaintext.model_validate({
            'pt': decrypter.update(bytes(ct)) + decrypter.finalize(),
            'aad': ct.aad
        })

    async def encrypt(
        self,
        pt: bytes | Plaintext,
        algorithm: IAlgorithm | None = None
    ) -> EncryptionResult:
        if isinstance(pt, bytes):
            pt = Plaintext(pt=pt)
        algorithm = algorithm or self.default_algorithm()
        mode, iv = algorithm.get_initialization_vector()
        cipher = Cipher(AES(self._key), mode)
        encrypter = cipher.encryptor()
        if algorithm.supports_aad():
            encrypter.authenticate_additional_data(b'') # type: ignore
        ct = encrypter.update(bytes(pt)) + encrypter.finalize()
        return EncryptionResult.model_validate({
            'aad': b'',
            'alg': algorithm.name,
            'kid': self.kid,
            'ct': ct,
            'iv': iv,
            'tag': encrypter.tag
        })

    async def sign( # type: ignore
        self,
        message: bytes,
        algorithm: HMACSigning,
    ) -> bytes:
        return hmac.digest(self._key, message, algorithm.digestmod)

    async def verify( # type: ignore
        self,
        signature: bytes,
        message: bytes,
        algorithm: HMACSigning,
        using: str | None = None
    ) -> bool:
        h = hmac.new(self._key, message, algorithm.digestmod)
        return hmac.compare_digest(h.digest(), signature)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import hashlib
from typing import Any
from typing import ClassVar
from typing import Literal

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric.utils import encode_dss_signature

import aiopki
from aiopki.algorithms import EllipticCurveSigning
from aiopki.types import IAlgorithm
from aiopki.utils import b64decode_int
from aiopki.utils import b64encode
from aiopki.utils import b64encode_int
from aiopki.utils import bytes_to_number
from aiopki.utils import normalize_ec_signature
from .basejwk import BaseJWK


class EllipticCurvePublicKey(BaseJWK):
    curves: ClassVar[dict[str, str]] = {
        'secp256k1': 'P-256K',
        'secp256r1': 'P-256',
        'secp384r1': 'P-384',
        'secp521r1': 'P-512',
    }
    kty: Literal['EC']
    crv: str # type: ignore
    x: str
    y: str
    _public_numbers: ec.EllipticCurvePublicNumbers

    @property
    def public(self): # pragma: no cover
        return self

    @property
    def public_key(self) -> ec.EllipticCurvePublicKey:
        return self._public_numbers.public_key()

    @property
    def thumbprint(self) -> str:
        message = self.public.model_dump_json(include={'crv', 'kty', 'x', 'y'})
        return b64encode(hashlib.sha256(str.encode(message)).digest(), encoder=bytes.decode)

    @classmethod
    def preprocess_key(cls, key: ec.EllipticCurvePublicKey | Any, **values: Any) -> dict[str, Any]:
        if not isinstance(key, ec.EllipticCurvePublicKey):
            raise NotImplementedError
        numbers = key.public_numbers()
        return {
            'kty': 'EC',
            'use': 'sig',
            'crv': EllipticCurvePublicKey.curves[key.curve.name],
            'x': b64encode_int(numbers.x),
            'y': b64encode_int(numbers.y),
        }

    def can_use(self, algorithm: IAlgorithm) -> bool:
        if not isinstance(algorithm, EllipticCurveSigning):
            return False
        return algorithm.curve == self.crv

    def model_post_init(self, *args: Any, **kwargs: Any) -> None:
        self._public_numbers = ec.EllipticCurvePublicNumbers(
            curve=aiopki.curve(self.crv),
            x=b64decode_int(self.x),
            y=b64decode_int(self.y)
        )

    async def verify(
        self,
        signature: bytes,
        message: bytes,
        algorithm: IAlgorithm,
        using: str | None = None
    ) -> bool:
        n = (self.public_key.curve.key_size + 7) // 8
        try:
            self.public_key.verify(
                signature=encode_dss_signature(
                    bytes_to_number(signature[:n]),
                    bytes_to_number(signature[n:]),
                ),
                data=message,
                signature_algorithm=ec.ECDSA(algorithm.get_digest_algorithm())
            )
            return True
        except InvalidSignature:
            return False

    def process_signature(self, signature: bytes) -> bytes:
        return normalize_ec_signature(
            l=(self.public_key.curve.key_size + 7) // 8,
            sig=signature
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import hashlib
from typing import Any

from google.cloud.kms import KeyManagementServiceAsyncClient
from google.cloud.kms import CryptoKeyVersion as GoogleCryptoKeyVersion
from google.cloud.kms import CryptoKey
from cryptography.hazmat.primitives.serialization import load_pem_public_key

from aiopki import CryptoKeyVersion
from aiopki import VersionedCryptoKey
from aiopki.lib import JSONWebKey
from aiopki.types import IAlgorithm
from aiopki.utils import b64encode
from aiopki.utils import crc32c
from aiopki.utils import crc32c_check
from .client import get_credential


AlgorithmEnum = GoogleCryptoKeyVersion.CryptoKeyVersionAlgorithm
PurposeEnum = CryptoKey.CryptoKeyPurpose
VersionStateEnum = GoogleCryptoKeyVersion.CryptoKeyVersionState


ALGORITHM_MAPPING: dict[AlgorithmEnum, str] = {
    AlgorithmEnum.HMAC_SHA1                 : 'HS1',
    AlgorithmEnum.HMAC_SHA224               : 'HS224',
    AlgorithmEnum.HMAC_SHA256               : 'HS256',
    AlgorithmEnum.HMAC_SHA384               : 'HS384',
    AlgorithmEnum.HMAC_SHA512               : 'HS512',
    AlgorithmEnum.EC_SIGN_P256_SHA256       : 'ES256',
    AlgorithmEnum.RSA_SIGN_PKCS1_4096_SHA256: 'RS256',
}


class GoogleKey(VersionedCryptoKey):

    def model_post_init(self, _: Any) -> None:
        super().model_post_init(_)
        if self.backend != 'cloudkms.googleapis.com':
            raise ValueError(f"Invalid backend: {self.backend}")

    async def discover(self):
        client = KeyManagementServiceAsyncClient(credentials=await get_credential())
        key = await client.get_crypto_key(name=self.name) # type: ignore
        async for version in await client.list_crypto_key_versions(parent=self.name): # type: ignore
            is_enabled = version.state == VersionStateEnum.ENABLED
            thumbprint = None
            public = None
            if key.purpose in {PurposeEnum.MAC, PurposeEnum.ENCRYPT_DECRYPT}:
                thumbprint = b64encode(
                    hashlib.sha256(str.encode(version.name)).digest(),
                    encoder=bytes.decode
                )
            else:
                response = await client.get_public_key(name=version.name) # type: ignore
                if not response.name == version.name:
                    raise ValueError(
                        f"Version mismatch: {response.name} != {version.name}"
                    )
                if not crc32c_check(int(response.pem_crc32c), response.pem.encode('utf-8')): # type: ignore
                    raise ValueError("Invalid response checksum")
                public = JSONWebKey.model_validate({
                    'key': load_pem_public_key(str.encode(response.pem, 'utf-8')) # type: ignore
                })
                thumbprint = public.thumbprint
            params = client.parse_crypto_key_version_path(version.name)
            self.add_version(
                name=params['crypto_key_version'],
                alg=ALGORITHM_MAPPING[version.algorithm],
                enabled=is_enabled,
                thumbprint=thumbprint,
                public=public
            )
            if not self.has_default():
                self.default_version = thumbprint

    async def sign(
        self,
        version: CryptoKeyVersion,
        message: bytes,
        algorithm: IAlgorithm,
    ) -> bytes:
        client = KeyManagementServiceAsyncClient(credentials=await get_credential())
        name = client.crypto_key_version_path(
            crypto_key_version=version.name,
            **self.annotations # type: ignore
        )
        if version.alg in {'HS1','HS224', 'HS256', 'HS384', 'HS512'}:
            response = await client.mac_sign(name=name, data=message) # type: ignore
            sig = response.mac
        else:
            digest = algorithm.digest(message)
            response = await client.asymmetric_sign( # type: ignore
                request={
                    'name': name,
                    'digest': {algorithm.get_digest_name(): digest},
                    'digest_crc32c': crc32c(digest)
                }
            )
            if not response.verified_digest_crc32c:
                raise ValueError("The request sent to the server was corrupted in-transit.")
            if not response.name == name:
                raise ValueError("The request sent to the server was corrupted in-transit.")
            if not response.signature_crc32c == crc32c(response.signature):
                raise ValueError(
                    "The response received from the server was "
                    "corrupted in-transit."
                )
            sig = version.process_signature(response.signature)

        return sig

    async def verify(
        self,
        version: CryptoKeyVersion,
        signature: bytes,
        message: bytes,
        algorithm: IAlgorithm
    ) -> bool:
        client = KeyManagementServiceAsyncClient(credentials=await get_credential())
        name = client.crypto_key_version_path(
            crypto_key_version=version.name,
            **self.annotations # type: ignore
        )
        if version.alg in {'HS1','HS224', 'HS256', 'HS384', 'HS512'}:
            response = await client.mac_verify( # type: ignore
                name=name,
                data=message,
                mac=signature
            )
            result = response.success
        elif version.public is not None:
            result = await version.public.verify(
                signature=signature,
                message=message,
                algorithm=algorithm,
            )
        else:
            raise NotImplementedError
        return result


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from canonical import HTTPResourceLocator
from canonical import ResourceName
from google.cloud.kms import KeyManagementServiceClient
from google.cloud.secretmanager import SecretManagerServiceClient

from .googlekey import GoogleKey
from .googlesecret import GoogleSecret


SUPPORTED_SERVICES: set[str] = {
    'cloudkms.googleapis.com',
    'secretmanager.googleapis.com',
}


def parse(name: ResourceName | HTTPResourceLocator) -> GoogleSecret | GoogleKey | None:
    if not isinstance(name, ResourceName):
        return None
    if name.service not in SUPPORTED_SERVICES:
        return None

    match str(name.service):
        case 'cloudkms.googleapis.com':
            params = KeyManagementServiceClient.parse_crypto_key_version_path(name.relname)
            if not params:
                params = KeyManagementServiceClient.parse_crypto_key_path(name.relname)
            if not params:
                return None
            return GoogleKey.model_validate({ # type: ignore
                'backend': 'cloudkms.googleapis.com',
                'name': str(name.relname),
                'default_version': params.get('crypto_key_version') or '__default__',
                'annotations': params
            })
        case 'secretmanager.googleapis.com':
            params = SecretManagerServiceClient.parse_secret_version_path(name.relname)
            if not params:
                params = SecretManagerServiceClient.parse_secret_path(name.relname)
            if not params:
                return None
            return GoogleSecret.model_validate({
                'backend': 'cloudkms.googleapis.com',
                'name': SecretManagerServiceClient.secret_path(
                    project=params['project'],
                    secret=params['secret']
                ),
                'default_version': params.get('secret_version') or '__default__',
                'annotations': params
            })
        case _:
            raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest

from aiopki.ext import jose


A1: str = (
    'eyJhbGciOiJSU0EtT0FFUCIsImVuYyI6IkEyNTZHQ00ifQ.OKOawDo13gRp2ojaHV'
    '7LFpZcgV7T6DVZKTyKOMTYUmKoTCVJRgckCL9kiMT03JGeipsEdY3mx_etLbbWSrF'
    'r05kLzcSr4qKAq7YN7e9jwQRb23nfa6c9d-StnImGyFDbSv04uVuxIp5Zms1gNxKK'
    'K2Da14B8S4rzVRltdYwam_lDp5XnZAYpQdb76FdIKLaVmqgfwX7XWRxv2322i-vDx'
    'RfqNzo_tETKzpVLzfiwQyeyPGLBIO56YJ7eObdv0je81860ppamavo35UgoRdbYaB'
    'coh9QcfylQr66oc6vFWXRcZ_ZT2LawVCWTIy3brGPi6UklfCpIMfIjf7iGdXKHzg.'
    '48V1_ALb6US04U3b.5eym8TW_c8SuK0ltJ3rpYIzOeDQz7TALvtu6UG9oMo4vpzs9'
    'tX_EFShS8iB7j6jiSdiwkIr3ajwQzaBtQD_A.XFBoMYUZodetZdvTiFvSkQ'
)


@pytest.mark.parametrize("value", [A1])
def test_parse_rfc(value: str):
    jwe = jose.parse(value)


@pytest.mark.asyncio
async def test_can_not_verify_signature():
    jwe = jose.parse(A1)
    with pytest.raises(TypeError):
        await jwe.verify(NotImplemented)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest

from aiopki.ext import jose
from .const import SUPPORTED_KEYS


@pytest.mark.asyncio
@pytest.mark.parametrize("kid", SUPPORTED_KEYS)
async def test_sign_bytes(jwks: jose.JWKS, kid: str):
    k = jwks.get(kid)
    assert k.alg is not None
    t = await jose.sign(k, k.alg, b'Hello world!')
    o = jose.parse(t)
    p = o.payload(factory=bytes)
    assert await o.verify(k)
    assert p == b'Hello world!'


@pytest.mark.asyncio
@pytest.mark.parametrize("kid", SUPPORTED_KEYS)
async def test_sign_claims(jwks: jose.JWKS, kid: str):
    k = jwks.get(kid)
    assert k.alg is not None
    t = await jose.sign(k, k.alg, {'sub': 'username'})
    o = jose.parse(t)
    p = o.payload(factory=jose.JWT.model_validate)
    assert await o.verify(k)
    assert p.sub == 'username'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest

import aiopki
from aiopki.ext.jose import JWKS
from .const import ENCRYPTION_KEYS


PLAINTEXT: bytes = b'Hello world!'


@pytest.mark.parametrize("kid,alg", ENCRYPTION_KEYS)
@pytest.mark.asyncio
async def test_jwk_encrypt_decrypt(jwks: JWKS, kid: str, alg: str):
    k = jwks.get(kid)
    a = aiopki.algorithms.get(alg)
    ct = await k.encrypt(PLAINTEXT, a)
    pt = await k.decrypt(ct, a)
    assert bytes(pt) == PLAINTEXT


@pytest.mark.parametrize("kid,alg", ENCRYPTION_KEYS)
@pytest.mark.asyncio
async def test_jwks_encrypt_decrypt(jwks: JWKS, kid: str, alg: str):
    a = aiopki.algorithms.get(alg)
    ct = await jwks.encrypt(PLAINTEXT, a)
    pt = await jwks.decrypt(ct, a)
    assert bytes(pt) == PLAINTEXT


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from aiopki.types import IVerifier


class JWE(pydantic.BaseModel):
    ciphertext: bytes
    iv: bytes
    payload: str | None = None

    def encode(self) -> bytes:
        """Encodes the JWS/JWE."""
        raise NotImplementedError

    async def verify(self, verifier: IVerifier) -> bool:
        raise TypeError("JWE does not have a signature.")


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest

import pydantic
from aiopki import CryptoKeyType


class Model(pydantic.BaseModel):
    key: CryptoKeyType


@pytest.mark.parametrize("name", [
    '//cloudkms.googleapis.com/projects/unimatrixdev/locations/europe-west4/keyRings/local/cryptoKeys/aes256gcm',
    '//cloudkms.googleapis.com/projects/unimatrixdev/locations/europe-west4/keyRings/local/cryptoKeys/aes256gcm/cryptoKeyVersions/3'
])
def test_parse_resource_name(name: str):
    obj = Model.model_validate({'key': name})
    assert isinstance(obj.key, CryptoKeyType)